;;; php-mode.el --- major mode for editing PHP code

;; Author:       Vincent DEBOUT <deboutv@free.fr>
;; Maintainer:	Vincent DEBOUT <deboutv@free.fr>
;; Keywords:	languages php
;; WWW:		http://deboutv.free.fr/lisp/php/

;; History
;;
;; $Log: php-mode.el,v $
;; Revision 1.2  2006/07/23 15:47:18  vincent
;; Add Template, Fontification, Indentation
;;
;; Revision 1.1.1.1  2006/07/14 10:16:37  vincent
;; First import
;;

(defconst php-version "0.0.1"
  "PHP Mode version number.")

(defconst php-time-stamp "2006-07-23"
  "PHP Mode time stamp for last update.")

(require 'font-lock)
(require 'assoc)
(require 'compile)
(require 'easymenu)
(require 'hippie-exp)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Variables
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Tempalte variable

(defvar php-template-menu (list '("Control Structures"
				  ["break" php-template-break t]
				  ["continue" php-template-continue t]
				  ;["declare" php-template-declare t]
				  ["do-while" php-template-do-while t]
				  ["else" php-template-else t]
				  ["elseif" php-template-elseif t]
				  ["for" php-template-for t]
				  ["foreach" php-template-foreach t]
				  ["if" php-template-if t]
				  ["include" php-template-include t]
				  ["include_once" php-template-include-once t]
				  ["require" php-template-require t]
				  ["require_once" php-template-require-once t]
				  ["return" php-template-return t]
				  ["switch" php-template-switch t]
				  ["while" php-template-while t])
				'("File System"
				  ("File System (Basename --> File_get_contents)"
				   ["basename" php-template-basename t]
				   ["chgrp" php-template-chgrp t]
				   ["chmod" php-template-chmod t]
				   ["chown" php-template-chown t]
				   ["clearstatcache" php-template-clearstatcache t]
				   ["copy" php-template-copy t]
				   ["delete" php-template-delete t]
				   ["dirname" php-template-dirname t]
				   ["disk_free_space" php-template-disk-free-space t]
				   ["disk_total_space" php-template-disk-total-space t]
				   ["diskfreespace" php-template-diskfreespace t]
				   ["fclose" php-template-fclose t]
				   ["feof" php-template-feof t]
				   ["fflush" php-template-fflush t]
				   ["fgetc" php-template-fgetc t]
				   ["fgetcsv" php-template-fgetcsv t]
				   ["fgets" php-template-fgets t]
				   ["fgetss" php-template-fgetss t]
				   ["file_exists" php-template-file-exists t]
				   ["file_get_contents" php-template-file-get-contents t])
				  ("File System (File_put_contents --> Fseek)"
				   ["file_put_contents" php-template-file-put-contents t]
				   ["file" php-template-file t]
				   ["fileatime" php-template-fileatime t]
				   ["filectime" php-template-filectime t]
				   ["filegroup" php-template-filegroup t]
				   ["fileinode" php-template-fileinode t]
				   ["filemtime" php-template-filemtime t]
				   ["fileowner" php-template-fileowner t]
				   ["fileperms" php-template-fileperms t]
				   ["filesize" php-template-filesize t]
				   ["filetype" php-template-filetype t]
				   ["flock" php-template-flock t]
				   ["fnmatch" php-template-fnmatch t]
				   ["fopen" php-template-fopen t]
				   ["fpassthru" php-template-fpassthru t]
				   ["fputcsv" php-template-fputcsv t]
				   ["fputs" php-template-fwrite t]
				   ["fread" php-template-fread t]
					;  ["fscanf" php-template-fscanf t]
				   ["fseek" php-template-fseek t])
				  ("File System (Fstat --> Pathinfo)"
				   ["fstat" php-template-fstat t]
				   ["ftell" php-template-ftell t]
				   ["ftruncate" php-template-ftruncate t]
				   ["fwrite" php-template-fwrite t]
				   ["glob" php-template-glob t]
				   ["is_dir" php-template-is-dir t]
				   ["is_executable" php-template-is-executable t]
				   ["is_file" php-template-is-file t]
				   ["is_link" php-template-is-link t]
				   ["is_readable" php-template-is-readable t]
				   ["is_uploaded_file" php-template-is-uploaded-file t]
				   ["is_writable" php-template-is-writable t]
				   ["is_writeable" php-template-is-writable t]
				   ["link" php-template-link t]
				   ["linkinfo" php-template-linkinfo t]
				   ["lstat" php-template-lstat t]
				   ["mkdir" php-template-mkdir t]
				   ["move_uploaded_file" php-template-move-uploaded-file t]
				   ["parse_ini_file" php-template-parse-ini-file t]
				   ["pathinfo" php-template-pathinfo t])
				  ("File System (Pclose --> Unlink)"
				   ["pclose" php-template-pclose t]
				   ["popen" php-template-popen t]
				   ["readfile" php-template-readfile t]
				   ["readlink" php-template-readlink t]
				   ["realpath" php-template-realpath t]
				   ["rename" php-template-rename t]
				   ["rewind" php-template-rewind t]
				   ["rmdir" php-template-rmdir t]
				   ;["set_file_buffer" php-template-set-file-buffer t]
				   ["stat" php-template-stat t]
				   ["symlink" php-template-symlink t]
				   ["tempnam" php-template-tempnam t]
				   ["tmpfile" php-template-tmpfile t]
				   ["touch" php-template-touch t]
				   ["umask" php-template-umask t]
				   ["unlink" php-template-unlink t]))
				'("MySQL"
				  ("MySQL (Aff --> Field_flags)"
				   ["mysql_affected_rows" php-template-mysql-affected-rows t]
				   ["mysql_change_user" php-template-mysql-change-user t]
				   ["mysql_client_encoding" php-template-mysql-client-encoding t]
				   ["mysql_close" php-template-mysql-close t]
				   ["mysql_connect" php-template-mysql-connect t]
				   ["mysql_create_db" php-template-mysql-create-db t]
				   ["mysql_data_seek" php-template-mysql-data-seek t]
				   ["mysql_db_name" php-template-mysql-db-name t]
				   ["mysql_db_query" php-template-mysql-db-query t]
				   ["mysql_drop_db" php-template-mysql-drop-db t]
				   ["mysql_errno" php-template-mysql-errno t]
				   ["mysql_error" php-template-mysql-error t]
				   ["mysql_escape_string" php-template-mysql-escape-string t]
				   ["mysql_fetch_array" php-template-mysql-fetch-array t]
				   ["mysql_fetch_assoc" php-template-mysql-fetch-assoc t]
				   ["mysql_fetch_field" php-template-mysql-fetch-field t]
				   ["mysql_fetch_lengths" php-template-mysql-fetch-lengths t]
				   ["mysql_fetch_object" php-template-mysql-fetch-object t]
				   ["mysql_fetch_row" php-template-mysql-fetch-row t]
				   ["mysql_field_flags" php-template-mysql-field-flags t])
				  ("MySQL (Field_lan --> Ping"
				   ["mysql_field_len" php-template-mysql-field-len t]
				   ["mysql_field_name" php-template-mysql-field-name t]
				   ["mysql_field_seek" php-template-mysql-field-seek t]
				   ["mysql_field_table" php-template-mysql-field-table t]
				   ["mysql_field_type" php-template-mysql-field-type t]
				   ["mysql_free_result" php-template-mysql-free-result t]
				   ["mysql_get_client_info" php-template-mysql-get-client-info t]
				   ["mysql_get_host_info" php-template-mysql-get-host-info t]
				   ["mysql_get_proto_info" php-template-mysql-get-proto-info t]
				   ["mysql_get_server_info" php-template-mysql-get-server-info t]
				   ["mysql_info" php-template-mysql-info t]
				   ["mysql_insert_id" php-template-mysql-insert-id t]
				   ["mysql_list_dbs" php-template-mysql-list-dbs t]
				   ["mysql_list_fields" php-template-mysql-list-fields t]
				   ["mysql_list_processes" php-template-mysql-list-processes t]
				   ["mysql_list_tables" php-template-mysql-list-tables t]
				   ["mysql_num_fields" php-template-mysql-num-fields t]
				   ["mysql_num_rows" php-template-mysql-num-rows t]
				   ["mysql_pconnect" php-template-mysql-pconnect t]
				   ["mysql_ping" php-template-mysql-ping t])
				  ("MySQL (Query --> Unbuff)"
				   ["mysql_query" php-template-mysql-query t]
				   ["mysql_real_escape_string" php-template-mysql-real-escape-string t]
				   ["mysql_result" php-template-mysql-result t]
				   ["mysql_select_db" php-template-mysql-select-db t]
				   ["mysql_stat" php-template-mysql-stat t]
				   ["mysql_tablename" php-template-mysql-tablename t]
				   ["mysql_thread_id" php-template-mysql-thread-id t]
				   ["mysql_unbuffered_query" php-template-mysql-unbuffered-query t]))
				["Insert Header" php-template-header t]
				["Insert Footer" php-template-footer t]
				["Insert Date" php-template-insert-date t]
				["Modify Date" php-template-modify t]
				)
  "The Template menu")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Customization
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgroup php nil
  "Customizations for PHP mode."
  :prefix "php-"
  :group 'languages)

(defgroup php-mode nil
  "Customizations for modes."
  :group 'php)

(defcustom php-electric-mode t
  "*Non-nil enables electrification (automatic template generation).
If nil, template generators can still be invoked through key bindings and
menu.  Is indicated in the modeline by \"/e\" after the mode name and can be
toggled by `\\[php-electric-mode]'."
  :type 'boolean
  :group 'php-mode)

(defcustom php-stutter-mode t
  "*Non-nil enables stuttering.
Is indicated in the modeline by \"/s\" after the mode name and can be toggled
by `\\[php-stutter-mode]'."
  :type 'boolean
  :group 'php-mode)

(defcustom php-indent-tabs-mode nil
  "*Non-nil means indentation can insert tabs.
Overrides local variable `indent-tabs-mode'."
  :type 'boolean
  :group 'php-mode)

(defgroup php-menu nil
  "Customizations for menues."
  :group 'php)

(defcustom php-index-menu t
  "*Non-nil means add an index menu for a source file when loading.
Note that the index menu scans a file
when it is opened."
  :type 'boolean
  :group 'php-menu)

(defcustom php-source-file-menu t
  "*Non-nil means add a menu of all source files in current directory."
  :type 'boolean
  :group 'php-menu)

(defgroup php-template nil
  "Template group."
  :group 'php)

(defcustom php-conditions-in-parenthesis t
  "*Non-nil means place parenthesis around condition expressions."
  :type 'boolean
  :group 'php-template)

(defcustom php-include-in-parenthesis t
  "*Non-nil means place parenthesis around include file."
  :type 'boolean
  :group 'php-template)

(defcustom php-add-fclose-with-fopen t
  "*Non-nil means that fclose statement will be added during the fopen template."
  :type 'boolean
  :group 'php-template)

(defcustom php-add-mysql-close-when-connect t
  "*Non-nil means that mysql_close statement will be added during the mysql_connect template."
  :type 'boolean
  :group 'php-template)

(defgroup php-style nil
  "Customizations for coding styles."
  :group 'php)

(defcustom php-basic-offset 4
  "*Amount of basic offset used for indentation.
This value is used by + and - symbols in `php-offsets-alist'."
  :type 'integer
  :group 'php-style)

(defgroup php-header nil
  "Customizations for file header."
  :group 'php-template)

(defcustom php-file-header "\
/**
 * <description string>
 *
 *
 * Created: <date>
 * Last update: <date>
 *
 * @link <link string>
 * @copyright <copyright>
 * @author <author>
 * @package <package string>
 * @version <version string>
 */

<cursor>
"
  "*String or file to insert as file header.
If the string specifies an existing file name, the contents of the file is
inserted, otherwise the string itself is inserted as file header.
Type `C-j' for newlines.
If the header contains RCS keywords, they may be written as <RCS>Keyword<RCS>
if the header needs to be version controlled.

The following keywords for template generation are supported:
  <filename>    : replaced by the name of the buffer
  <author>      : replaced by the user name and email address
                  \(`user-full-name',`mail-host-address', `user-mail-address')
  <login>       : replaced by user login name (`user-login-name')
  <company>     : replaced by contents of option `php-company-name'
  <date>        : replaced by the current date
  <year>        : replaced by the current year
  <copyright>   : replaced by copyright string (`php-copyright-string')
  <cursor>      : final cursor position."
  :type 'string
  :group 'php-header)

(defcustom php-file-footer ""
  "*String or file to insert as file footer.
If the string specifies an existing file name, the contents of the file is
inserted, otherwise the string itself is inserted as file footer (i.e. at
the end of the file).
Type `C-j' for newlines.
The same keywords as in option `php-file-header' can be used."
  :type 'string
  :group 'php-header)

(defcustom php-company-name ""
  "*Name of company to insert in file header.
See option `php-file-header'."
  :type 'string
  :group 'php-header)

(defcustom php-copyright-string ""
  "*Copyright string to insert in file header.
Can be multi-line string (type `C-j' for newline) and contain other file
header keywords (see option `php-file-header')."
  :type 'string
  :group 'php-header)

(defcustom php-date-format "%Y-%m-%d"
  "*Specifies the date format to use in the header.
This string is passed as argument to the command `format-time-string'.
For more information on format strings, see the documentation for the
`format-time-string' command (C-h f `format-time-string')."
  :type 'string
  :group 'php-header)

(defcustom php-modify-date-prefix-string " * Last update: "
  "*Prefix string of modification date in PHP file header.
If actualization of the modification date is called (menu,
`\\[php-template-modify]'), this string is searched and the rest
of the line replaced by the current date."
  :type 'string
  :group 'php-header)

(defcustom php-modify-date-on-saving t
  "*Non-nil means update the modification date when the buffer is saved.
Calls function `\\[php-template-modify]').

NOTE: Activate the new setting in a PHP buffer by using the menu entry
      \"Activate Options\"."
  :type 'boolean
  :group 'php-header)

(defgroup php-misc nil
  "Miscellaneous customizations."
  :group 'php)

(defcustom php-intelligent-tab t
  "*Non-nil means `TAB' does indentation, word completion and tab insertion.
That is, if preceding character is part of a word then complete word,
else if not at beginning of line then insert tab,
else if last command was a `TAB' or `RET' then dedent one step,
else indent current line (i.e. `TAB' is bound to `php-electric-tab').
If nil, TAB always indents current line (i.e. `TAB' is bound to
`indent-according-to-mode').

NOTE: Activate the new setting in a PHP buffer by using the menu entry
      \"Activate Options\"."
  :type 'boolean
  :group 'php-misc)

(defcustom php-word-completion-in-minibuffer t
  "*Non-nil enables word completion in minibuffer (for template prompts).

NOTE: Activate the new setting by restarting Emacs."
  :type 'boolean
  :group 'php-misc)

(defcustom php-word-completion-case-sensitive nil
  "*Non-nil means word completion using `TAB' is case sensitive.
That is, `TAB' completes words that start with the same letters and case.
Otherwise, case is ignored."
  :type 'boolean
  :group 'php-misc)

(defun php-custom-set (variable value &rest functions)
  "Set variables as in `custom-set-default' and call FUNCTIONS afterwards."
  (if (fboundp 'custom-set-default)
      (custom-set-default variable value)
    (set-default variable value))
  (while functions
    (when (fboundp (car functions)) (funcall (car functions)))
    (setq functions (cdr functions))))

(defun php-customize ()
  "Call the customize function with `php' as argument."
  (interactive)
  (customize-browse 'php))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Menu tools functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun php-menu-split (list title)
  "Split menu LIST into several submenues, if number of
elements > `php-menu-max-size'."
  (if (> (length list) php-menu-max-size)
      (let ((remain list)
	    (result '())
	    (sublist '())
	    (menuno 1)
	    (i 0))
	(while remain
	  (setq sublist (cons (car remain) sublist))
	  (setq remain (cdr remain))
	  (setq i (+ i 1))
	  (if (= i php-menu-max-size)
	      (progn
		(setq result (cons (cons (format "%s %s" title menuno)
					 (nreverse sublist)) result))
		(setq i 0)
		(setq menuno (+ menuno 1))
		(setq sublist '()))))
	(and sublist
	     (setq result (cons (cons (format "%s %s" title menuno)
				      (nreverse sublist)) result)))
	(nreverse result))
    list))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Index menu
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defconst php-imenu-generic-expression
  '(
    ("Functions"
     "^\\s-*\\(function\\)\\s-+\\(\\(\\w\\|\\s_\\)+\\)"
     2)
    ("Class"
     "^\\s-*\\(class\\)\\s-+\\(\\(\\w\\|\\s_\\)+\\)"
     2)
    )
  "Imenu generic expression for PHP Mode.  See `imenu-generic-expression'.")

(defun php-index-menu-init ()
  "Initialize index menu."
  (set (make-local-variable 'imenu-case-fold-search) t)
  (set (make-local-variable 'imenu-generic-expression)
       php-imenu-generic-expression)
  (when (and php-index-menu (fboundp 'imenu))
    (if (or (not (boundp 'font-lock-maximum-size))
	    (> font-lock-maximum-size (buffer-size)))
	(imenu-add-to-menubar "Index")
      (message "Scanning buffer for index...buffer too big"))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Source file menu
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar php-sources-menu nil)



;; Create the source menu
(defun php-add-source-files-menu ()
  "Scan directory for all PHP source files and generate menu.
The directory of the current source file is scanned."
  (interactive)
  (message "Scanning directory for source files ...")
  (let ((newmap (current-local-map))
	(file-list (php-get-source-files))
	menu-list found)
    ;; Create list for menu
    (setq found nil)
    (while file-list
      (setq found t)
      (setq menu-list (cons (vector (car file-list)
				   (list 'find-file (car file-list)) t)
			   menu-list))
      (setq file-list (cdr file-list)))
    (setq menu-list (php-menu-split menu-list "Sources"))
    (when found (setq menu-list (cons "--" menu-list)))
    (setq menu-list (cons ["*Rescan*" php-add-source-files-menu t] menu-list))
    (setq menu-list (cons "Sources" menu-list))
    ;; Create menu
    (easy-menu-add menu-list)
    (easy-menu-define php-sources-menu newmap
		      "PHP source files menu" menu-list))
  (message ""))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; PHP menu
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun php-create-mode-menu ()
  "Create PHP Mode menu."
  `("PHP"
    ,(append
      '("Templates")
      php-template-menu)
    ("Indent"
     ["Line"			indent-according-to-mode :keys "C-c C-i C-l"]
     ["Region"			php-indent-region (mark)]
     ["Buffer"			php-indent-buffer :keys "C-c C-i C-b"])
    ["Show Messages"		php-show-messages t]
    "--"
    ["Version"			php-version t]
    "--"
    ("Options"
     ("Mode"
      ["Electric Mode"
       (progn (customize-set-variable 'php-electric-mode
				      (not php-electric-mode))
	      (php-mode-line-update))
       :style toggle :selected php-electric-mode :keys "C-c C-m C-e"]
      ["Stutter Mode"
       (progn (customize-set-variable 'php-stutter-mode
				      (not php-stutter-mode))
	      (php-mode-line-update))
       :style toggle :selected php-stutter-mode :keys "C-c C-m C-s"]
      ["Indent Tabs Mode"
       (progn (customize-set-variable 'php-indent-tabs-mode
				      (not php-indent-tabs-mode))
	      (setq indent-tabs-mode php-indent-tabs-mode))
       :style toggle :selected php-indent-tabs-mode]
      "--"
      ["Customize Group..." (customize-group 'php-mode) t])
     ("Menu"
      ["Index Menu"
       (customize-set-variable 'php-index-menu
			       (not php-index-menu))
       :style toggle :selected php-index-menu]
      ["Source Menu"
       (customize-set-variable 'php-source-file-menu
			       (not php-source-file-menu))
       :style toggle :selected php-source-file-menu]
      "--"
      ["Customize Group..." (customize-group 'php-menu) t])
     ("Template"
      ["Condition in parenthesis"
       (customize-set-variable 'php-conditions-in-parenthesis
			       (not php-conditions-in-parenthesis))
       :style toggle :selected php-conditions-in-parenthesis]
      ["Parenthesis in include"
       (customize-set-variable 'php-include-in-parenthesis
			       (not php-include-in-parenthesis))
       :style toggle :selected php-include-in-parenthesis]
      ["fopen -> fclose"
       (customize-set-variable 'php-add-fclose-with-fopen
			       (not php-add-fclose-with-fopen))
       :style toggle :selected php-add-fclose-with-fopen]
      ["mysql_connect -> mysql_close"
       (customize-set-variable 'php-add-mysql-close-when-connect
			       (not php-add-mysql-close-when-connect))
       :style toggle :selected php-add-mysql-close-when-connect]
      "--"
      ["Customize Group..." (customize-group 'php-template) t])
     ("Style"
      ["Indent offset..."
       (customize-option 'php-basic-offset) t]
      "--"
      ["Customize Group..." (customize-group 'php-style) t])
     ("Header"
      ["Header template..."
       (customize-option 'php-file-header) t]
      ["Footer template..."
       (customize-option 'php-file-footer) t]
      ["Company..."
       (customize-option 'php-company-name) t]
      ["Copyright..."
       (customize-option 'php-copyright-string) t]
      ["Date format..."
       (customize-option 'php-date-format) t]
      ["Modify date prefix..."
       (customize-option 'php-modify-date-prefix-string) t]
      ["Modify date on saving"
       (customize-set-variable 'php-modify-date-on-saving
			       (not php-modify-date-on-saving))
       :style toggle :selected php-modify-date-on-saving]
      "--"
      ["Customize Group..." (customize-group 'php-header) t])
     ("Miscellaneous"
      ["Use Intelligent Tab"
       (progn (customize-set-variable 'php-intelligent-tab
				      (not php-intelligent-tab))
	      (php-activate-customizations))
       :style toggle :selected php-intelligent-tab]
      ["Word Completion in Minibuffer"
       (progn (customize-set-variable 'php-word-completion-in-minibuffer
				      (not php-word-completion-in-minibuffer))
	      (message "Activate new setting by saving options and restarting Emacs"))
       :style toggle :selected php-word-completion-in-minibuffer]
      ["Completion is case sensitive"
       (customize-set-variable 'php-word-completion-case-sensitive
			       (not php-word-completion-case-sensitive))
       :style toggle :selected php-word-completion-case-sensitive]
      "--"
      ["Customize Group..." (customize-group 'php-misc) t])
     "--"
     ["Save Options" customize-save-customized t]
     ["Activate Options" php-activate-customizations t]
     ["Browse Options..." php-customize t])))

(defvar php-mode-menu-list (php-create-mode-menu)
  "PHP Mode menu.")

(defun php-update-mode-menu ()
  "Update PHP Mode menu."
  (interactive)
  (easy-menu-remove php-mode-menu-list)
  (setq php-mode-menu-list (php-create-mode-menu))
  (easy-menu-add php-mode-menu-list)
  (easy-menu-define php-mode-menu php-mode-map
		    "Menu keymap for PHP Mode." php-mode-menu-list))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Variables
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar php-menu-max-size 20
  "*Specifies the maximum size of a menu before splitting it into submenues.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Progress reporting
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar php-progress-interval 1
  "*Interval used to update progress status during long operations.
If a number, percentage complete gets updated after each interval of
that many seconds.  To inhibit all messages, set this option to nil.")

(defvar php-progress-info nil
  "Array variable for progress information: 0 begin, 1 end, 2 time.")

(defun php-update-progress-info (string pos)
  "Update progress information."
  (when (and php-progress-info (not noninteractive)
	     (< php-progress-interval
		(- (nth 1 (current-time)) (aref php-progress-info 2))))
    (message (concat string "... (%2d%s)")
	     (/ (* 100 (- pos (aref php-progress-info 0)))
		(- (aref php-progress-info 1)
		   (aref php-progress-info 0))) "%")
    (aset php-progress-info 2 (nth 1 (current-time)))))

(defmacro php-point (position)
  "Return the value of point at certain commonly referenced POSITIONs.
POSITION can be one of the following symbols:

bol  -- beginning of line
eol  -- end of line
bod  -- beginning of defun
boi  -- back to indentation
eoi  -- last whitespace on line
ionl -- indentation of next line
iopl -- indentation of previous line
bonl -- beginning of next line
bopl -- beginning of previous line

This function does not modify point or mark."
  (or (and (eq 'quote (car-safe position))
	   (null (cddr position)))
      (error "ERROR:  Bad buffer position requested: %s" position))
  (setq position (nth 1 position))
  `(let ((here (point)))
     ,@(cond
	((eq position 'bol)  '((beginning-of-line)))
	((eq position 'eol)  '((end-of-line)))
	((eq position 'bod)  '((save-match-data
				 (php-beginning-of-defun))))
	((eq position 'boi)  '((back-to-indentation)))
	((eq position 'eoi)  '((end-of-line) (skip-chars-backward " \t")))
	((eq position 'bonl) '((forward-line 1)))
	((eq position 'bopl) '((forward-line -1)))
	((eq position 'iopl)
	 '((forward-line -1)
	   (back-to-indentation)))
	((eq position 'ionl)
	 '((forward-line 1)
	   (back-to-indentation)))
	(t (error "ERROR:  Unknown buffer position requested: %s" position))
	)
     (prog1
	 (point)
       (goto-char here))
     ;; workaround for an Emacs18 bug -- blech! Well, at least it
     ;; doesn't hurt for v19
     ,@nil
     ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; File/directory manipulation
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun php-directory-files (directory &optional full match)
  "Call `directory-files' if DIRECTORY exists, otherwise generate error
message."
  (if (not (file-directory-p directory))
      (php-warning-when-idle "No such directory: \"%s\"" directory)
    (let ((dir (directory-files directory full match)))
      (setq dir (delete "." dir))
      (setq dir (delete ".." dir))
      dir)))

(defun php-get-source-files (&optional full directory)
  "Get list of PHP source files in DIRECTORY or current directory."
  (let ((mode-alist auto-mode-alist)
	filename-regexp)
    ;; create regular expressions for matching file names
    (setq filename-regexp "\\`[^.].*\\(")
    (while mode-alist
      (when (eq (cdar mode-alist) 'php-mode)
	(setq filename-regexp
	      (concat filename-regexp (caar mode-alist) "\\|")))
      (setq mode-alist (cdr mode-alist)))
    (setq filename-regexp
	  (concat (substring filename-regexp 0
			     (string-match "\\\\|$" filename-regexp)) "\\)"))
    ;; find files
    (php-directory-files
     (or directory default-directory) full filename-regexp)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Messages reporting
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar php-warnings nil
  "Warnings to tell the user during start up.")

(defun php-run-when-idle (secs repeat function)
  "Wait until idle, then run FUNCTION."
  (if (fboundp 'start-itimer)
      (start-itimer "php-mode" function secs repeat t)
;    (run-with-idle-timer secs repeat function)))
    ;; explicitely activate timer (necessary when Emacs is already idle)
    (aset (run-with-idle-timer secs repeat function) 0 nil)))

(defun php-warning-when-idle (&rest args)
  "Wait until idle, then print out warning STRING and beep."
  (if noninteractive
      (php-warning (apply 'format args) t)
    (unless php-warnings
      (php-run-when-idle .1 nil 'php-print-warnings))
    (setq php-warnings (cons (apply 'format args) php-warnings))))

(defun php-warning (string &optional nobeep)
  "Print out warning STRING and beep."
  (message (concat "WARNING:  " string))
  (unless (or nobeep noninteractive) (beep)))

(defun php-print-warnings ()
  "Print out messages in variable `php-warnings'."
  (let ((no-warnings (length php-warnings)))
    (setq php-warnings (nreverse php-warnings))
    (while php-warnings
      (message (concat "WARNING:  " (car php-warnings)))
      (setq php-warnings (cdr php-warnings)))
    (beep)
    (when (> no-warnings 1)
      (message "WARNING:  See warnings in message buffer (type `C-c M-m')."))))

(defun php-show-messages ()
  "Get *Messages* buffer to show recent messages."
  (interactive)
  (display-buffer " *Message-Log*"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Update the date field when saving file
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun php-write-file-hooks-init ()
  "Add/remove hooks when buffer is saved."
  (if php-modify-date-on-saving
      (add-hook 'local-write-file-hooks 'php-template-modify-noerror)
    (remove-hook 'local-write-file-hooks 'php-template-modify-noerror))
  (make-local-variable 'after-save-hook)
  (add-hook 'after-save-hook 'php-add-modified-file))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Global variable
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar php-mode-map nil
  "Keymap for PHP Mode.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Variables
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Regular expression used internally to recognize labels in switch
;;; statements.
(defvar php-switch-label-regexp (purecopy "case[ \t'/(]\\|default[ \t]*:"))

(defvar php-label-offset -2
  "*Offset of PHP label lines and case statements relative to usual indentation.")

(defvar php-indent-level 4
  "*Indentation of PHP statements with respect to containing block.")

(defvar php-brace-offset 0
  "*Extra indentation for braces, compared with other text in same context.")

(defvar php-argdecl-indent 5
  "*Indentation level of declarations of PHP function arguments.")

(defvar php-continued-statement-offset 2
  "*Extra indent for lines not starting new statements.")

(defvar php-continued-brace-offset 0
  "*Extra indent for substatements that start with open-braces.")

(defvar php-brace-imaginary-offset 0
  "*Imagined indentation of a PHP open brace that actually follows a statement.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun php-comment-indent ()
  (if (looking-at "^/\\*")
      0				;Existing comment at bol stays there.
    (let ((opoint (point)))
      (save-excursion
	(beginning-of-line)
	(cond ((looking-at "[ \t]*}[ \t]*\\($\\|/\\*\\)")
	       ;; A comment following a solitary close-brace
	       ;; should have only one space.
	       (search-forward "}")
	       (1+ (current-column)))
	      ((or (looking-at "^#[ \t]*endif[ \t]*")
		   (looking-at "^#[ \t]*else[ \t]*"))
	       7)			;2 spaces after #endif
	      ((progn
		 (goto-char opoint)
		 (skip-chars-backward " \t")
		 (and (= comment-column 0) (bolp)))
	       ;; If comment-column is 0, and nothing but space
	       ;; before the comment, align it at 0 rather than 1.
	       0)
	      (t
	       (max (1+ (current-column))	;Else indent at comment column
		    comment-column)))))))	; except leave at least one space.

(defun php-indent-line-2 ()
  "Indent current line as PHP code.
Return the amount the indentation changed by."
  (let ((indent (php-calculate-indent nil))
	beg shift-amt
	(case-fold-search nil)
	(pos (- (point-max) (point))))
    (beginning-of-line)
    (setq beg (point))
    (cond ((eq indent nil)
	   (setq indent (current-indentation)))
	  ((eq indent t)
	   (setq indent (php-calculate-indent-within-comment)))
	  ((looking-at "[ \t]*#")
	   (setq indent 0))
	  (t
	   (skip-chars-forward " \t")
	   (if (listp indent) (setq indent (car indent)))
	   (cond ((or (looking-at php-switch-label-regexp)
		      (and (looking-at "[A-Za-z]")
			   (save-excursion
			     (forward-sexp 1)
			     (looking-at ":"))))
		  (setq indent (max 1 (+ indent php-label-offset))))
		 ((and (looking-at "else\\b")
		       (not (looking-at "else\\s_")))
		  (setq indent (save-excursion
				 (php-backward-to-start-of-if)
				 (current-indentation))))
		 ((and (looking-at "}[ \t]*else\\b")
		       (not (looking-at "}[ \t]*else\\s_")))
		  (setq indent (save-excursion
				 (forward-char)
				 (backward-sexp)
				 (php-backward-to-start-of-if)
				 (current-indentation))))
		 ((and (looking-at "while\\b")
		       (not (looking-at "while\\s_"))
		       (save-excursion
			 (php-backward-to-start-of-do)))
		  ;; This is a `while' that ends a do-while.
		  (setq indent (save-excursion
				 (php-backward-to-start-of-do)
				 (current-indentation))))
		 ((= (following-char) ?})
		  (setq indent (- indent php-indent-level)))
		 ((= (following-char) ?{)
		  (setq indent (+ indent php-brace-offset))))))
    (skip-chars-forward " \t")
    (setq shift-amt (- indent (current-column)))
    (if (zerop shift-amt)
	(if (> (- (point-max) pos) (point))
	    (goto-char (- (point-max) pos)))
      (delete-region beg (point))
      (indent-to indent)
      ;; If initial point was within line's indentation,
      ;; position after the indentation.  Else stay at same point in text.
      (if (> (- (point-max) pos) (point))
	  (goto-char (- (point-max) pos))))
    shift-amt))

(defun php-indent-line ()
  "Indent the current line as PHP code.  Returns the amount of
indentation change."
  (interactive)
  (let* (indent current-line)
    (save-excursion
      (beginning-of-line)
      (setq current-line (point-marker))
      ;(while (and (not (eq (point-marker) (beginning-of-buffer)))
		  ;(
	(forward-line -1);)
      (if (eq current-line (point-marker))
	  (setq indent 0)
	(setq indent (current-indentation))
	(if (and (re-search-forward "^\\s-*\\/\\*" (php-point 'eol) t) (not (re-search-forward "\\*\\/\\s-*$" (php-point 'eol) t)))
	    (setq indent (+ indent 1)))
	(if (and (re-search-forward "\\*\\/\\s-*$" (php-point 'eol) t) (not (re-search-forward "^\\s-*\\/\\*" (php-point 'eol) t)))
	    (setq indent (- indent 1)))
	(when (and (re-search-forward "\\({\\|(\\)\\s-*$" (php-point 'eol) t ) (not (re-search-forward "^\\(}\\|)\\)\\s-*" (php-point 'eol) t)))
	  (setq indent (+ indent php-basic-offset)))))
    (delete-region (php-point 'bol) (php-point 'boi))
    (beginning-of-line)
    (when (re-search-forward "^\\(}\\|)\\)\\s-*" (php-point 'eol) t)
      (setq indent (- indent php-basic-offset)))
    (indent-to indent)
    indent))

(defun php-indent-region (beg end column)
  "Indent region as PHP code.
Adds progress reporting to `indent-region'."
  (interactive "r\nP")
  (when php-progress-interval
    (setq php-progress-info (vector (count-lines (point-min) beg)
				     (count-lines (point-min) end) 0)))
  (indent-region beg end column)
  (when php-progress-interval (message "Indenting...done"))
  (setq php-progress-info nil))

(defun php-indent-buffer ()
  "Indent whole buffer as PHP code.
Calls `indent-region' for whole buffer and adds progress reporting."
  (interactive)
  (php-indent-region (point-min) (point-max) nil))

(defun php-calculate-indent (&optional parse-start)
  "Return appropriate indentation for current line as C code.
In usual case returns an integer: the column to indent to.
Returns nil if line starts inside a string, t if in a comment."
  (save-excursion
    (beginning-of-line)
    (let ((indent-point (point))
	  (case-fold-search nil)
	  state
	  containing-sexp)
      (if parse-start
	  (goto-char parse-start)
	(beginning-of-defun))
      (while (< (point) indent-point)
	(setq parse-start (point))
	(setq state (parse-partial-sexp (point) indent-point 0))
	(setq containing-sexp (car (cdr state))))
      (cond ((or (nth 3 state) (nth 4 state))
	     ;; return nil or t if should not change this line
	     (nth 4 state))
	    ((null containing-sexp)
	     ;; Line is at top level.  May be data or function definition,
	     ;; or may be function argument declaration.
	     ;; Indent like the previous top level line
	     ;; unless that ends in a closeparen without semicolon,
	     ;; in which case this line is the first argument decl.
	     (goto-char indent-point)
	     (skip-chars-forward " \t")
	     (if (= (following-char) ?{)
		 0   ; Unless it starts a function body
	       (php-backward-to-noncomment (or parse-start (point-min)))
	       ;; Look at previous line that's at column 0
	       ;; to determine whether we are in top-level decls
	       ;; or function's arg decls.  Set basic-indent accordingly.
	       (let ((basic-indent
		      (save-excursion
			(re-search-backward "^[^ \^L\t\n#]" nil 'move)
			(let (comment lim)
			  ;; Recognize the DEFUN macro in Emacs.
			  (if (save-excursion
				;; Move down to the (putative) argnames line.
				(while (and (not (eobp))
					    (not (looking-at " *[({}#/]")))
				  (forward-line 1))
				;; Go back to the DEFUN, if it is one.
				(condition-case nil
				    (backward-sexp 1)
				  (error))
				(beginning-of-line)
				(looking-at "DEFUN\\b"))
			      php-argdecl-indent
			    (if (and (looking-at "\\sw\\|\\s_")
				     ;; This is careful to stop at the first
				     ;; paren if we have
				     ;; int foo Proto ((int, int));
				     (looking-at "[^\"\n=(]*(")
				     (progn
				       (goto-char (1- (match-end 0)))
				       ;; Skip any number of paren-groups.
				       ;; Consider typedef int (*fcn) (int);
				       (while (= (following-char) ?\()
					 (setq lim (point))
					 (condition-case nil
					     (forward-sexp 1)
					   (error))
					 (skip-chars-forward " \t\f"))
				       ;; Have we reached something
				       ;; that shows this isn't a function
				       ;; definition?
				       (and (< (point) indent-point)
					    (not (memq (following-char)
						       '(?\, ?\;)))))
				     ;; Make sure the "function decl" we found
				     ;; is not inside a comment.
				     (progn
				       ;; Move back to the `(' starting arglist
				       (goto-char lim)
				       (beginning-of-line)
				       (while (and (not comment)
						   (search-forward "/*" lim t))
					 (setq comment
					       (not (search-forward "*/" lim t))))
				       (not comment)))
				php-argdecl-indent 0))))))
		 basic-indent)))

;; 		 ;; Now add a little if this is a continuation line.
;; 		 (+ basic-indent (if (or (bobp)
;; 					 (memq (preceding-char) '(?\) ?\; ?\}))
;; 					 ;; Line with zero indentation
;; 					 ;; is probably the return-type
;; 					 ;; of a function definition,
;; 					 ;; so following line is function name.
;; 					 (= (current-indentation) 0))
;;				     0 php-continued-statement-offset))

	    ((/= (char-after containing-sexp) ?{)
	     ;; line is expression, not statement:
	     ;; indent to just after the surrounding open.
	     (goto-char (1+ containing-sexp))
	     (current-column))
	    (t
	     ;; Statement level.  Is it a continuation or a new statement?
	     ;; Find previous non-comment character.
	     (goto-char indent-point)
	     (php-backward-to-noncomment containing-sexp)
	     ;; Back up over label lines, since they don't
	     ;; affect whether our line is a continuation.
	     (while (or (eq (preceding-char) ?\,)
			(and (eq (preceding-char) ?:)
			     (or (eq (char-after (- (point) 2)) ?\')
				 (memq (char-syntax (char-after (- (point) 2)))
				       '(?w ?_)))))
	       (if (eq (preceding-char) ?\,)
		   (progn (forward-char -1)
			  (php-backward-to-start-of-continued-exp containing-sexp)))
	       (beginning-of-line)
	       (php-backward-to-noncomment containing-sexp))
	     ;; Check for a preprocessor statement or its continuation lines.
	     ;; Move back to end of previous non-preprocessor line,
	     ;; or possibly beginning of buffer.
	     (let ((found (point)) stop)
	       (while (not stop)
		 (beginning-of-line)
		 (cond ((bobp)
			(setq found (point)
			      stop t))
		       ((save-excursion (forward-char -1)
					(= (preceding-char) ?\\))
			(forward-char -1))
		       ;; This line is not preceded by a backslash.
		       ;; So either it starts a preprocessor command
		       ;; or any following continuation lines
		       ;; should not be skipped.
		       ((= (following-char) ?#)
			(forward-char -1)
			(setq found (point)))
		       (t (setq stop t))))
	       (goto-char found))
	     ;; Now we get the answer.
	     (if (and (not (memq (preceding-char) '(0 ?\, ?\; ?\} ?\{)))
		      ;; But don't treat a line with a close-brace
		      ;; as a continuation.  It is probably the
		      ;; end of an enum type declaration.
		      (save-excursion
			(goto-char indent-point)
			(skip-chars-forward " \t")
			(not (= (following-char) ?}))))
		 ;; This line is continuation of preceding line's statement;
		 ;; indent  php-continued-statement-offset  more than the
		 ;; previous line of the statement.
		 (progn
		   (php-backward-to-start-of-continued-exp containing-sexp)
		   (+ php-continued-statement-offset (current-column)
		      (if (save-excursion (goto-char indent-point)
					  (skip-chars-forward " \t")
					  (eq (following-char) ?{))
			  php-continued-brace-offset 0)))
	       ;; This line starts a new statement.
	       ;; Position following last unclosed open.
	       (goto-char containing-sexp)
	       ;; Is line first statement after an open-brace?
	       (or
		 ;; If no, find that first statement and indent like it.
		 (save-excursion
		   (forward-char 1)
		   (let ((colon-line-end 0))
		     (while (progn (skip-chars-forward " \t\n")
				   (looking-at "#\\|/\\*\\|case[ \t\n'/(].*:\\|[a-zA-Z0-9_$]*:"))
		       ;; Skip over comments and labels following openbrace.
		       (cond ((= (following-char) ?\#)
			      (forward-line 1))
			     ((= (following-char) ?\/)
			      (forward-char 2)
			      (search-forward "*/" nil 'move))
			     ;; case or label:
			     (t
			      (save-excursion (end-of-line)
					      (setq colon-line-end (point)))
			      (search-forward ":"))))
		     ;; The first following code counts
		     ;; if it is before the line we want to indent.
		     (and (< (point) indent-point)
			  (- 
			   (if (> colon-line-end (point))
			       (- (current-indentation) php-label-offset)
			     (current-column))
			   ;; If prev stmt starts with open-brace, that
			   ;; open brace was offset by php-brace-offset.
			   ;; Compensate to get the column where
			   ;; an ordinary statement would start.
			   (if (= (following-char) ?\{) php-brace-offset 0)))))
		 ;; If no previous statement,
		 ;; indent it relative to line brace is on.
		 (php-calculate-indent-after-brace))))))))

(defun php-calculate-indent-within-comment (&optional after-star)
  "Return the indentation amount for line inside a block comment.
Optional arg AFTER-STAR means, if lines in the comment have a leading star,
return the indentation of the text that would follow this star."
  (let (end star-start)
    (save-excursion
      (beginning-of-line)
      (skip-chars-forward " \t")
      (setq star-start (= (following-char) ?\*))
      (skip-chars-backward " \t\n")
      (setq end (point))
      (beginning-of-line)
      (skip-chars-forward " \t")
      (if after-star
	  (and (looking-at "\\*")
	       (re-search-forward "\\*[ \t]*")))
      (and (re-search-forward "/\\*[ \t]*" end t)
	   star-start
	   (not after-star)
	   (goto-char (1+ (match-beginning 0))))
      (if (and (looking-at "[ \t]*$") (= (preceding-char) ?\*))
	  (1+ (current-column))
	(current-column)))))

(defun php-backward-to-noncomment (lim)
  (let (stop)
    (while (not stop)
      (skip-chars-backward " \t\n\f" lim)
      (if (and (>= (point) (+ 2 lim))
	       (save-excursion
		 (forward-char -2)
		 (looking-at "\\*/")))
	  (search-backward "/*" lim 'move)
	(setq stop (or (<= (point) lim)
		       (save-excursion
			 (while (progn
				  (beginning-of-line)
				  (eq ?\\ (char-after (- (point) 2))))
			   (forward-char -1)
			   (beginning-of-line))
			 (skip-chars-forward " \t")
			 (not (looking-at "#")))))
	(or stop (beginning-of-line))))))

(defun php-backward-to-start-of-continued-exp (lim)
  (if (memq (preceding-char) '(?\) ?\"))
      (forward-sexp -1))
  (beginning-of-line)
  (if (<= (point) lim)
      (goto-char (1+ lim)))
  (skip-chars-forward " \t"))

(defun php-backward-to-start-of-if (&optional limit)
  "Move to the start of the last \"unbalanced\" `if'."
  (or limit (setq limit (save-excursion (beginning-of-defun) (point))))
  (let ((if-level 1)
	(case-fold-search nil))
    (while (and (not (bobp)) (not (zerop if-level)))
      (backward-sexp 1)
      (cond ((and (looking-at "else\\b")
		  (not (looking-at "else\\s_")))
	     (setq if-level (1+ if-level)))
	    ((and (looking-at "if\\b")
		  (not (looking-at "if\\s_")))
	     (setq if-level (1- if-level)))
	    ((< (point) limit)
	     (setq if-level 0)
	     (goto-char limit))))))

(defun php-backward-to-start-of-do (&optional limit)
  "If point follows a `do' statement, move to beginning of it and return t.
Otherwise return nil and don't move point."
  (or limit (setq limit (save-excursion (beginning-of-defun) (point))))
  (let ((first t)
	(startpos (point))
	(done nil))
    (while (not done)
      (let ((next-start (point)))
	(condition-case nil
	    ;; Move back one token or one brace or paren group.
	    (backward-sexp 1)
	  ;; If we find an open-brace, we lose.
	  (error (setq done 'fail)))
	(if done
	    nil
	  ;; If we reached a `do', we win.
	  (if (looking-at "do\\b")
	      (setq done 'succeed)
	    ;; Otherwise, if we skipped a semicolon, we lose.
	    ;; (Exception: we can skip one semicolon before getting
	    ;; to a the last token of the statement, unless that token
	    ;; is a close brace.)
	    (if (save-excursion
		  (forward-sexp 1)
		  (or (and (not first) (= (preceding-char) ?}))
		      (search-forward ";" next-start t
				      (if (and first
					       (/= (preceding-char) ?}))
					  2 1))))
		(setq done 'fail)
	      (setq first nil)
	      ;; If we go too far back in the buffer, we lose.
	      (if (< (point) limit)
		  (setq done 'fail)))))))
    (if (eq done 'succeed)
	t
      (goto-char startpos)
      nil)))

(defun php-calculate-indent-after-brace ()
  "Return the proper PHP indent for the first line after an open-brace.
This function is called with point before the brace."
  ;; For open brace in column zero, don't let statement
  ;; start there too.  If php-indent-level is zero,
  ;; use php-brace-offset + php-continued-statement-offset instead.
  ;; For open-braces not the first thing in a line,
  ;; add in php-brace-imaginary-offset.
  (+ (if (and (bolp) (zerop php-indent-level))
	 (+ php-brace-offset php-continued-statement-offset)
       php-indent-level)
     ;; Move back over whitespace before the openbrace.
     ;; If openbrace is not first nonwhite thing on the line,
     ;; add the c-brace-imaginary-offset.
     (progn (skip-chars-backward " \t")
	    (if (bolp) 0 php-brace-imaginary-offset))
     ;; If the openbrace is preceded by a parenthesized exp,
     ;; move to the beginning of that;
     ;; possibly a different line
     (progn
       (if (eq (preceding-char) ?\))
	   (forward-sexp -1))
       ;; Get initial indentation of the line we are on.
       (current-indentation))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun php-function-name (prefix string &optional postfix)
  "Generate a Lisp function name.
PREFIX, STRING and optional POSTFIX are concatenated by '-' and spaces in
STRING are replaced by `-' and substrings are converted to lower case."
  (let ((name prefix))
    (while (string-match "\\(\\w+\\)\\s-*\\(.*\\)" string)
      (setq name
	    (concat name "-" (downcase (substring string 0 (match-end 1)))))
      (setq string (substring string (match-beginning 2))))
    (when postfix (setq name (concat name "-" postfix)))
    (intern name)))

(defvar php-mode-hook nil
  "*Hook called by `php-mode'.")

(defvar php-constants nil
  "List of PHP constants.")

(defvar php-constants-regexp nil
  "Regexp for PHP constants.")

(defconst php-01-constants
  '(;; core constants
    "__LINE__" "__FILE__"
    "PHP_OS" "PHP_VERSION"
    "TRUE" "FALSE" "NULL"
    "E_ERROR" "E_NOTICE" "E_PARSE" "E_WARNING" "E_ALL"
    "E_USER_ERROR" "E_USER_WARNING" "E_USER_NOTICE"
    "DEFAULT_INCLUDE_PATH" "PEAR_INSTALL_DIR" "PEAR_EXTENSION_DIR"
    "PHP_BINDIR" "PHP_LIBDIR" "PHP_DATADIR" "PHP_SYSCONFDIR"
    "PHP_LOCALSTATEDIR" "PHP_CONFIG_FILE_PATH")
  "PHP constants.")

(defvar php-keywords nil
  "List of PHP keywords.")

(defvar php-keywords-regexp nil
  "Regexp for PHP keywords.")

(defconst php-01-keywords
  ;; "class", "new" and "extends" get special treatment
  ;; "case" and  "default" get special treatment elsewhere
  '("and" "as" "break" "continue" "declare" "do" "echo" "else" "elseif"
    "endfor" "endforeach" "endif" "endswitch" "endwhile" "exit"
    "extends" "for" "foreach" "global" "if" "include" "include_once"
    "next" "or" "require" "require_once" "return" "static" "switch"
    "then" "var" "while" "xor" "private" "throw" "catch" "try"
    "instanceof" "catch all" "finally")
  "PHP keywords.")

(defconst php-identifier
  (eval-when-compile
    '"[a-zA-Z\_\x7f-\xff][a-zA-Z0-9\_\x7f-\xff]*")
  "Characters in a PHP identifier.")

(defvar php-types nil
  "List of PHP types.")

(defvar php-types-regexp nil
  "Regexp for PHP types.")

(defconst php-01-types
  '("array" "bool" "boolean" "char" "const" "double" "float"
    "int" "integer" "long" "mixed" "object" "real"
    "string")
  "PHP types.")

(defvar php-superglobals nil
  "List of PHP superglobals.")

(defvar php-superglobals-regexp nil
  "Regexp for PHP superglobals.")

(defconst php-01-superglobals
  '("_GET" "_POST" "_COOKIE" "_SESSION" "_ENV" "GLOBALS"
    "_SERVER" "_FILES" "_REQUEST")
  "PHP superglobal variables.")

(defvar php-functions nil
  "List of PHP functions.")

(defvar php-functions-regexp nil
  "Regexp for PHP functions.")

(defconst php-01-functions
  '("basename" "chgrp" "chmod" "chown" "clearstatcache" "copy" "delete"
    "dirname" "disk_free_space" "disk_total_space" "diskfreespace" "fclose"
    "feof" "fflush" "fgetc" "fgetcsv" "fgets" "fgetss" "file_exists"
    "file_get_contents" "file_put_contents" "file" "fileatime" "filectime"
    "filegroup" "fileinode" "filemtime" "fileowner" "fileperms" "filesize"
    "filetype" "flock" "fnmatch" "fopen" "fpassthru" "fputcsv" "fputs"
    "fread" "fscanf" "fseek" "fstat" "ftell" "ftruncate" "fwrite" "glob"
    "is_dir" "is_executable" "is_file" "is_link" "is_readable" "is_uploaded_file"
    "is_writable" "is_writeable" "link" "linkinfo" "lstat" "mkdir" "move_uploaded_file"
    "parse_ini_file" "pathinfo" "pclose" "popen" "readfile" "readlink" "realpath"
    "rename" "rewind" "rmdir" "set_file_buffer" "stat" "symlink" "tempnam"
    "tmpfile" "touch" "umask" "unlink" 
    "mysql_affected_rows" "mysql_change_user" "mysql_client_encoding"
    "mysql_close" "mysql_connect" "mysql_create_db" "mysql_data_seek"
    "mysql_db_name" "mysql_db_query" "mysql_drop_db" "mysql_errno"
    "mysql_error" "mysql_escape_string" "mysql_fetch_array" "mysql_fetch_assoc"
    "mysql_fetch_field" "mysql_fetch_lengths" "mysql_fetch_object"
    "mysql_fetch_row" "mysql_field_flags" "mysql_field_len"
    "mysql_field_name" "mysql_field_seek" "mysql_field_table" "mysql_field_type"
    "mysql_free_result" "mysql_get_client_info" "mysql_get_host_info"
    "mysql_get_proto_info" "mysql_get_server_info" "mysql_info" "mysql_insert_id"
    "mysql_list_dbs" "mysql_list_fields" "mysql_list_processes" "mysql_list_tables"
    "mysql_num_fields" "mysql_num_rows" "mysql_pconnect" "mysql_ping"
    "mysql_query" "mysql_real_escape_string" "mysql_result" "mysql_select_db"
    "mysql_stat" "mysql_tablename" "mysql_thread_id" "mysql_unbuffered_query")
  "PHP functions list")

(defun php-version ()
  "Echo the current version of PHP Mode in the minibuffer."
  (interactive)
  (message "PHP Mode %s (%s)" php-version php-time-stamp)
  (php-keep-region-active))

;; active regions
(defun php-keep-region-active ()
  "Do whatever is necessary to keep the region active in XEmacs.
Ignore byte-compiler warnings you might see."
  (and (boundp 'zmacs-region-stays)
       (setq zmacs-region-stays t)))

(defmacro php-prepare-search-1 (&rest body)
  "Enable case insensitive search and switch to syntax table that includes '_',
then execute BODY, and finally restore the old environment.  Used for
consistent searching."
  `(let ((case-fold-search t)		; case insensitive search
	 (current-syntax-table (syntax-table))
	 result
	 (restore-prog			; program to restore enviroment
	  '(progn
	     ;; restore syntax table
	     (set-syntax-table current-syntax-table))))
     ;; use extended syntax table
     (set-syntax-table php-mode-ext-syntax-table)
     ;; execute BODY safely
     (setq result
	   (condition-case info
	       (progn ,@body)
	     (error (eval restore-prog)	; restore environment on error
		    (error (cadr info))))) ; pass error up
     ;; restore environment
     (eval restore-prog)
     result))

(defmacro php-prepare-search-2 (&rest body)
  "Enable case insensitive search, switch to syntax table that includes '_',
and remove `intangible' overlays, then execute BODY, and finally restore the
old environment.  Used for consistent searching."
  `(let ((case-fold-search t)		; case insensitive search
	 (current-syntax-table (syntax-table))
	 result overlay-all-list overlay-intangible-list overlay
	 (restore-prog			; program to restore enviroment
	  '(progn
	     ;; restore syntax table
	     (set-syntax-table current-syntax-table)
	     ;; restore `intangible' overlays
	     (when (fboundp 'overlay-lists)
	       (while overlay-intangible-list
		 (overlay-put (car overlay-intangible-list) 'intangible t)
		 (setq overlay-intangible-list
		       (cdr overlay-intangible-list)))))))
     ;; use extended syntax table
     (set-syntax-table php-mode-ext-syntax-table)
     ;; remove `intangible' overlays
     (when (fboundp 'overlay-lists)
       (setq overlay-all-list (overlay-lists))
       (setq overlay-all-list
	     (append (car overlay-all-list) (cdr overlay-all-list)))
       (while overlay-all-list
	 (setq overlay (car overlay-all-list))
	 (when (memq 'intangible (overlay-properties overlay))
	   (setq overlay-intangible-list
		 (cons overlay overlay-intangible-list))
	   (overlay-put overlay 'intangible nil))
	 (setq overlay-all-list (cdr overlay-all-list))))
     ;; execute BODY safely
     (setq result
	   (condition-case info
	       (progn ,@body)
	     (error (eval restore-prog)	; restore environment on error
		    (error (cadr info))))) ; pass error up
     ;; restore environment
     (eval restore-prog)
     result))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Words to expand

(defun php-words-init ()
  "Initialize reserved words."
  (setq php-keywords php-01-keywords)
  (setq php-types php-01-types)
  (setq php-constants php-01-constants)
  (setq php-functions php-01-functions)
  (setq php-superglobals php-01-superglobals)
  (setq php-keywords-regexp (concat "\\<\\(" (regexp-opt php-keywords) "\\)\\>"))
  (setq php-types-regexp (concat "\\<\\(" (regexp-opt php-types) "\\)\\>"))
  (setq php-constants-regexp (concat "\\<\\(" (regexp-opt php-constants) "\\)\\>"))
  (setq php-functions-regexp (concat "\\<\\(" (regexp-opt php-functions) "\\)\\>"))
  (setq php-superglobals-regexp (concat "\\<\\(" (regexp-opt php-superglobals) "\\)\\>"))
  (php-abbrev-list-init))

(defvar php-abbrev-list nil
  "Predefined abbreviations for PHP.")

(defun php-abbrev-list-init ()
  (setq php-abbrev-list
	(append
	 (list nil) php-keywords
	 (list nil) php-types
	 (list nil) php-constants
	 (list nil) php-functions
	 (list nil) php-superglobals)))

(defvar php-expand-upper-case nil)

(defun php-try-expand-abbrev (old)
  "Try expanding abbreviations from `php-abbrev-list'."
  (unless old
    (he-init-string (he-dabbrev-beg) (point))
    (setq he-expand-list
	  (let ((abbrev-list php-abbrev-list)
		(sel-abbrev-list '()))
	    (while abbrev-list
	   ;   (if (stringp (car abbrev-list))
		;  (insert (concat " " (car abbrev-list))))
	      (when (or (not (stringp (car abbrev-list)))
			(string-match
			 (concat "^" he-search-string) (car abbrev-list)))
		(setq sel-abbrev-list
		      (cons (car abbrev-list) sel-abbrev-list)))
	      (setq abbrev-list (cdr abbrev-list)))
	    (nreverse sel-abbrev-list))))
  (while (and he-expand-list
	      (or (not (stringp (car he-expand-list)))
		  (he-string-member (car he-expand-list) he-tried-table t)))
    (unless (stringp (car he-expand-list))
      (setq php-expand-upper-case (car he-expand-list)))
    (setq he-expand-list (cdr he-expand-list)))
  (if (null he-expand-list)
      (progn (when old (he-reset-string))
	     nil)
    (he-substitute-string
     (if php-expand-upper-case
	 (upcase (car he-expand-list))
       (car he-expand-list))
     t)
    (setq he-expand-list (cdr he-expand-list))
    t))

;; initialize reserved words for PHP Mode
(php-words-init)

;; function for expanding abbrevs and dabbrevs
(defun php-expand-abbrev (arg))
(fset 'php-expand-abbrev (make-hippie-expand-function
			   '(try-expand-dabbrev
			     try-expand-dabbrev-all-buffers
			     php-try-expand-abbrev)))

;; function for expanding parenthesis
(defun php-expand-paren (arg))
(fset 'php-expand-paren (make-hippie-expand-function
			  '(try-expand-list
			    try-expand-list-all-buffers)))

;; Syntactic support functions:

(defun php-in-comment-p ()
  "Check if point is in a comment."
  (let (result (here (point)) begin)
    (save-excursion
      (beginning-of-line)
      (setq begin (point))
      (goto-char here)
      (if (or (or (re-search-backward "^\\s-*\#" begin t)
		  (re-search-backward "^\\s-*\/\/" begin t))
	      (php-in-comment-p2))
	  (setq result t)
	(setq result nil)))
    result))
  ;(eq (php-in-literal) 'comment))

(defun php-in-comment-p2 ()
  "Check if point is in a comment '/* */' type"
  (let (result start-comment (here (point)))
    (save-excursion
      (setq result t)
      (setq start-comment (re-search-backward "\\/\\*" nil t))
      (if (and start-comment (re-search-forward "\\*\\/" here t))
	  (setq result nil))
      (if (not start-comment)
	  (setq result nil)))
    result))

(defun php-in-string-p ()
  "Check if point is in a string."
  (eq (php-in-literal) 'string))

(defun php-in-quote-p ()
  "Check if point is in a quote ('x')."
  (or (and (> (point) (point-min))
	   (< (1+ (point)) (point-max))
	   (= (char-before (point)) ?\')
	   (= (char-after (1+ (point))) ?\'))
      (and (> (1- (point)) (point-min))
	   (< (point) (point-max))
	   (= (char-before (1- (point))) ?\')
	   (= (char-after (point)) ?\'))))

(defun php-in-literal ()
  "Determine if point is in a PHP literal."
  (save-excursion
    (let ((here (point))
	  start state)
      (beginning-of-line)
      (setq start (point))
      (goto-char here)
      (setq state (parse-partial-sexp start (point)))
      (cond
       ((nth 3 state) 'string)
       ((nth 4 state) 'comment)
       ((php-beginning-of-macro) 'pound)
       (t nil)))))

;; Macro definitions:

(defun php-beginning-of-macro (&optional lim)
  "Go to the beginning of a cpp macro definition (nicked from `cc-engine')."
  (let ((here (point)))
    (beginning-of-line)
    (while (eq (char-before (1- (point))) ?\\)
      (forward-line -1))
    (back-to-indentation)
    (if (and (<= (point) here)
	     (eq (char-after) ?#))
	t
      (goto-char here)
      nil)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Electrification
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defconst php-template-prompt-syntax "[^ =<>][^<>@.\n]*[^ =<>]"
  "Syntax of prompt inserted by template generators.")

(defvar php-template-invoked-by-hook nil
  "Indicates whether a template has been invoked by a hook or by key or menu.
Used for undoing after template abortion.")

(defun php-minibuffer-tab (&optional prefix-arg)
  "If preceding character is part of a word or a paren then hippie-expand,
else insert tab (used for word completion in PHP minibuffer)."
  (interactive "P")
  (cond
   ;; expand word
   ((= (char-syntax (preceding-char)) ?w)
    (let ((case-fold-search (not php-word-completion-case-sensitive))
	  (case-replace nil)
	  (hippie-expand-only-buffers
	   (or (and (boundp 'hippie-expand-only-buffers)
		    hippie-expand-only-buffers)
	       '(php-mode))))
      (php-expand-abbrev prefix-arg)))
   ;; expand parenthesis
   ((or (= (preceding-char) ?\() (= (preceding-char) ?\)))
    (let ((case-fold-search (not php-word-completion-case-sensitive))
	  (case-replace nil))
      (php-expand-paren prefix-arg)))
   ;; insert tab
   (t (insert-tab))))

;; correct different behavior of function `unread-command-events' in XEmacs
(defun php-character-to-event (arg))
(defalias 'php-character-to-event
  (if (fboundp 'character-to-event) 'character-to-event 'identity))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Syntax table

(defvar php-mode-syntax-table nil
  "Syntax table used in `php-mode' buffers.")

(defvar php-mode-ext-syntax-table nil
  "Syntax table extended by `_' used in `php-mode' buffers.")

(defun php-mode-syntax-table-init ()
  "Initialize `php-mode-syntax-table'."
  (setq php-mode-syntax-table (make-syntax-table))
  ;; define punctuation
  (modify-syntax-entry ?\# "< b"    php-mode-syntax-table)
  (modify-syntax-entry ?\$ "."    php-mode-syntax-table)
  (modify-syntax-entry ?\% "."    php-mode-syntax-table)
  (modify-syntax-entry ?\& "."    php-mode-syntax-table)
  (modify-syntax-entry ?\' "."    php-mode-syntax-table)
  (modify-syntax-entry ?\+ "."    php-mode-syntax-table)
  (modify-syntax-entry ?\. "."    php-mode-syntax-table)
  (modify-syntax-entry ?\: "."    php-mode-syntax-table)
  (modify-syntax-entry ?\; "."    php-mode-syntax-table)
  (modify-syntax-entry ?\< "."    php-mode-syntax-table)
  (modify-syntax-entry ?\= "."    php-mode-syntax-table)
  (modify-syntax-entry ?\> "."    php-mode-syntax-table)
  (modify-syntax-entry ?\\ "w"    php-mode-syntax-table)
  (modify-syntax-entry ?\| "."    php-mode-syntax-table)
  ;; define string
  (modify-syntax-entry ?\" "\""   php-mode-syntax-table)
  (modify-syntax-entry ?\' "\""   php-mode-syntax-table)
  ;; define underscore
  (modify-syntax-entry ?\_ "w"   php-mode-syntax-table)
  ;; a single hyphen is punctuation, but a double hyphen starts a comment
  (modify-syntax-entry ?\/ ". 14" php-mode-syntax-table)
  (modify-syntax-entry ?\* ". 23" php-mode-syntax-table)
  ;; and \n and \^M end a comment
  (modify-syntax-entry ?\n "> b"    php-mode-syntax-table)
  (modify-syntax-entry ?\^M "> b"   php-mode-syntax-table)
  ;; define parentheses to match
  (modify-syntax-entry ?\( "()"   php-mode-syntax-table)
  (modify-syntax-entry ?\) ")("   php-mode-syntax-table)
  (modify-syntax-entry ?\[ "(]"   php-mode-syntax-table)
  (modify-syntax-entry ?\] ")["   php-mode-syntax-table)
  (modify-syntax-entry ?\{ "(}"   php-mode-syntax-table)
  (modify-syntax-entry ?\} "){"   php-mode-syntax-table)
  ;; extended syntax table including '_' (for simpler search regexps)
  (setq php-mode-ext-syntax-table (copy-syntax-table php-mode-syntax-table))
  (modify-syntax-entry ?_ "w" php-mode-ext-syntax-table))

;; initialize syntax table for PHP Mode
(php-mode-syntax-table-init)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Mode map
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun php-mode-map-init ()
  "Initialize `php-mode-map'."
  (setq php-mode-map (make-sparse-keymap))
  ;; template key bindings
  ;;(define-key php-mode-map "\C-c\C-t"	   php-template-map)
  ;; mode specific key bindings
  (define-key php-mode-map "\C-c\C-m\C-e"  'php-electric-mode)
  (define-key php-mode-map "\C-c\C-m\C-s"  'php-stutter-mode)
  (define-key php-mode-map "\C-c\M-b"	   'php-beautify-region)
  (define-key php-mode-map "\C-c\C-b"	   'php-beautify-buffer)
  (define-key php-mode-map "\M-\C-\\"	   'php-indent-region)
  (define-key php-mode-map "\C-c\C-i\C-b"  'php-indent-buffer)
  (define-key php-mode-map "\C-cf"	   'php-fontify-buffer)
  (define-key php-mode-map "\C-c\M-m"	   'php-show-messages)
  (define-key php-mode-map "\C-c\C-h"	   'php-doc-mode)
  (define-key php-mode-map "\C-c\C-v"	   'php-version)
  (define-key php-mode-map "\M-\t"	   'insert-tab)
  ;; electric key bindings
  (define-key php-mode-map " "		   'php-electric-space)
  (when php-intelligent-tab
    (define-key php-mode-map "\t"	   'php-electric-tab))
  (define-key php-mode-map "\r"	           'php-electric-return)
  (define-key php-mode-map "*"             'php-electric-star)
  (define-key php-mode-map "?"             'php-electric-interrogation-point)
  (define-key php-mode-map "("		   'php-electric-open-bracket)
  (define-key php-mode-map ")"		   'php-electric-close-bracket)
  (define-key php-mode-map "}"		   'php-electric-close-bracket2)
  ;(define-key php-mode-map "'"		   'php-electric-quote)
  (define-key php-mode-map ";"		   'php-electric-semicolon))
  ;(define-key php-mode-map ","		   'php-electric-comma)
  ;(define-key php-mode-map "."		   'php-electric-period))

;; initialize mode map for PHP Mode
(php-mode-map-init)

;; define special minibuffer keymap for enabling word completion in minibuffer
;; (useful in template generator prompts)
(defvar php-minibuffer-local-map
  (let ((map (make-sparse-keymap)))
    (set-keymap-parent map minibuffer-local-map)
    (when php-word-completion-in-minibuffer
      (define-key map "\t" 'php-minibuffer-tab))
    map)
  "Keymap for minibuffer used in PHP Mode.")

;; set up electric character functions to work with
;; `delete-selection-mode' (Emacs) and `pending-delete-mode' (XEmacs)
(mapcar
 (function
  (lambda (sym)
    (put sym 'delete-selection t)	; for `delete-selection-mode' (Emacs)
    (put sym 'pending-delete t)))	; for `pending-delete-mode' (XEmacs)
 '(php-electric-space
   php-electric-tab
   php-electric-return
   php-electric-star
   php-electric-interrogation-point
   ;php-electric-dash
   php-electric-open-bracket
   php-electric-close-bracket
   php-electric-close-bracket2
   ;php-electric-quote
   php-electric-semicolon))
   ;php-electric-comma
   ;php-electric-period
   ;php-electric-equal))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Indentation commands

(defun php-electric-tab (&optional prefix-arg)
  "If preceding character is part of a word or a paren then hippie-expand,
else if right of non whitespace on line then insert tab,
else if last command was a tab or return then dedent one step or if a comment
toggle between normal indent and inline comment indent,
else indent `correctly'."
  (interactive "*P")
  (php-prepare-search-2
   (cond
    ;; expand word
    ((= (char-syntax (preceding-char)) ?w)
     (let ((case-fold-search (not php-word-completion-case-sensitive))
	   (case-replace nil)
	   (hippie-expand-only-buffers
	    (or (and (boundp 'hippie-expand-only-buffers)
		     hippie-expand-only-buffers)
		'(php-mode))))
       (php-expand-abbrev prefix-arg)))
    ;; expand parenthesis
    ((or (= (preceding-char) ?\() (= (preceding-char) ?\)))
     (let ((case-fold-search (not php-word-completion-case-sensitive))
	   (case-replace nil))
       (php-expand-paren prefix-arg)))
    ;; insert tab
    ((> (current-column) (current-indentation))
     (insert-tab))
    ;; toggle comment indent
    ((and (looking-at "//")
	  (or (eq last-command 'php-electric-tab)
	      (eq last-command 'php-electric-return)))
     (cond ((= (current-indentation) 0) ; no indent
	    (indent-to 1)
	    (indent-according-to-mode))
	   ((< (current-indentation) comment-column) ; normal indent
	    (indent-to comment-column)
	    (indent-according-to-mode))
	   (t				; inline comment indent
	    (kill-line -0))))
    ;; dedent
    ((and (>= (current-indentation) php-basic-offset)
	  (or (eq last-command 'php-electric-tab)
	      (eq last-command 'php-electric-return)))
     (backward-delete-char-untabify php-basic-offset nil))
    ;; indent line
    (t (indent-according-to-mode)))
   (setq this-command 'php-electric-tab)))

(defun php-electric-return ()
  "newline-and-indent or indent-new-comment-line if in comment and preceding
character is a space."
  (interactive)
  (if (php-in-comment-p2)
      (php-indent-new-comment-line)
    (newline-and-indent)))

(defun php-electric-space (count)
  "Expand abbreviations and self-insert space(s), do indent-new-comment-line
if in comment and past end-comment-column."
  (interactive "p")
  (cond ((php-in-comment-p)
	 (self-insert-command count)
	 (cond ((>= (current-column) (+ 2 end-comment-column))
		(backward-char 1)
		(skip-chars-backward "^ \t\n")
		(indent-new-comment-line)
		(skip-chars-forward "^ \t\n")
		(forward-char 1))
	       ((>= (current-column) end-comment-column)
		(indent-new-comment-line))
	       (t nil)))
	((or (and (>= (preceding-char) ?a) (<= (preceding-char) ?z))
	     (and (>= (preceding-char) ?A) (<= (preceding-char) ?Z)))
	 (php-prepare-search-1 (expand-abbrev))
	  ;(or (expand-abbrev) (php-fix-case-word -1)))
	 (self-insert-command count))
	(t (self-insert-command count))))

(defun php-electric-star (count)
  "/** start a comment paragraph"
  (interactive "p")
  (if (and php-stutter-mode (not (php-in-literal)))
      (if (= (preceding-char) ?\/)
	  (progn
	    (insert "*")
	    (message "Enter '*' for paragraph, 'SP' for commeting-out code")
	    (let ((next-input (read-char))
		  here)
	      (if (= next-input ?\*)
		  (progn
		    (insert "* \n")
		    (insert "* ")
		    (php-indent-line)
		    (end-of-line)
		    (setq here (point))
		    (insert "\n*/")
		    (php-indent-line)
		    (goto-char here))
		(insert " ")
		(setq here (point))
		(insert " */")
		(goto-char here))))
	(self-insert-command count))
    (self-insert-command count)))

(defun php-electric-interrogation-point (count)
  "Add '?>' and 'php' to the '<?' tag"
  (interactive "p")
  (let (here)
    (if (and php-stutter-mode (= count 1) (not (php-in-literal)))
	(if (= (preceding-char) ?\<)
	    (progn (insert "?php\n\n")
		   (setq here (point-marker))
		   (insert "\n\n?>\n")
		   (goto-char here))
	  (self-insert-command count))
      (self-insert-command count))))

(defun php-electric-semicolon (count)
  "Add a new line after a ';'"
  (interactive "p")
  (if (and php-stutter-mode (not (php-in-literal)))
      (progn
	(insert ";")
	(newline-and-indent))
    (self-insert-command count)))

(defun php-electric-open-bracket (count) 
  "'(' --> '(', '((' --> '[', '[(' --> '{'"
  (interactive "p")
  (if (and php-stutter-mode (= count 1) (not (php-in-literal)))
      (if (= (preceding-char) ?\()
	  (progn (delete-char -1) (insert-char ?\[ 1))
	(if (= (preceding-char) ?\[)
	    (progn (delete-char -1) (insert-char ?\{ 1))
	  (insert-char ?\( 1)))
    (self-insert-command count)))

(defun php-electric-close-bracket (count) 
  "')' --> ')', '))' --> ']', '])' --> '}'"
  (interactive "p")
  (if (and php-stutter-mode (= count 1) (not (php-in-literal)))
      (progn
	(if (= (preceding-char) ?\))
	    (progn (delete-char -1) (insert-char ?\] 1))
	  (if (= (preceding-char) ?\])
	      (progn (delete-char -1) (insert-char ?} 1) (save-excursion (php-indent-line)))
	    (insert-char ?\) 1)))
	(blink-matching-open))
    (self-insert-command count)))

(defun php-electric-close-bracket2 (count)
  "Automatic indent"
  (interactive "p")
  (if (and php-stutter-mode (not (php-in-literal)))
      (progn
	(insert "}")
	(save-excursion
	  (php-indent-line)))
    (self-insert-command count)))

(defun php-indent-new-comment-line ()
  "Add a new line, indent it and comment it"
  (interactive)
  (let ((here (point)) found)
    (setq found (re-search-backward "^\\s-*\\*" (php-point 'bol) t))
    (goto-char here)
    (newline-and-indent)
    (if found
	(insert "* "))))
  








;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  Enabling/disabling

(defun php-mode-line-update ()
  "Update the modeline string for PHP major mode."
  (setq mode-name (concat "PHP"
			  (and (or php-electric-mode php-stutter-mode) "/")
			  (and php-electric-mode "e")
			  (and php-stutter-mode "s")))
  (force-mode-line-update t))

(defun php-electric-mode (arg)
  "Toggle PHP electric mode.
Turn on if ARG positive, turn off if ARG negative, toggle if ARG zero or nil."
  (interactive "P")
  (setq php-electric-mode
	(cond ((or (not arg) (zerop arg)) (not php-electric-mode))
	      ((> arg 0) t) (t nil)))
  (php-mode-line-update))

(defun php-stutter-mode (arg)
  "Toggle PHP stuttering mode.
Turn on if ARG positive, turn off if ARG negative, toggle if ARG zero or nil."
  (interactive "P")
  (setq php-stutter-mode
	(cond ((or (not arg) (zerop arg)) (not php-stutter-mode))
	      ((> arg 0) t) (t nil)))
  (php-mode-line-update))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Fontification
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Set up font locking
(defconst php-font-lock-keywords-1
  (list
   ;; Fontify constants
   (cons
    (concat "\\<\\(" php-constants-regexp "\\)\\>")
    'font-lock-constant-face)

   ;; Fontify keywords
   (cons
    (concat "\\<\\(" php-keywords-regexp "\\)\\>")
    'font-lock-keyword-face)

   (cons
    (concat "\\<\\(" php-functions-regexp "\\)\\>")
    'font-lock-keyword-face)

   ;; Fontify keywords and targets, and case default tags.
   (list "\\<\\(break\\|case\\|continue\\)\\>[ \t]*\\(-?\\(?:\\sw\\|\\s_\\)+\\)?"
	 '(1 font-lock-keyword-face) '(2 font-lock-constant-face t t))
   ;; This must come after the one for keywords and targets.
   '(":" ("^[ \t]*\\(\\(?:\\sw\\|\\s_\\)+\\)[ \t]*:[ \t]*$"
	  (beginning-of-line) (end-of-line)
	  (1 font-lock-constant-face)))

   ;; treat 'print' as keyword only when not used like a function name
   '("\\<print\\s-*(" . default)
   '("\\<print\\>" . font-lock-keyword-face)

   ;; Fontify PHP tag
   '("<\\?\\(php\\)?" . font-lock-constant-face)
   '("\\?>" . font-lock-constant-face)

   ;; Fontify ASP-style tag
   '("<\\%\\(=\\)?" . font-lock-constant-face)
   '("\\%>" . font-lock-constant-face)

   )
  "Subdued level highlighting for PHP mode.")

(defconst php-font-lock-keywords-2
  (append
   php-font-lock-keywords-1
   (list

    ;; class declaration
    '("\\<\\(class\\|interface\\)[ \t]*\\(\\(?:\\sw\\|\\s_\\)+\\)?"
      (1 font-lock-keyword-face) (2 font-lock-type-face nil t))
    ;; handle several words specially, to include following word,
    ;; thereby excluding it from unknown-symbol checks later
    ;; FIX to handle implementing multiple
    ;; currently breaks on "class Foo implements Bar, Baz"
    '("\\<\\(new\\|extends\\|implements\\)\\s-+\\$?\\(\\(?:\\sw\\|\\s_\\)+\\)"
      (1 font-lock-keyword-face) (2 font-lock-type-face))

    ;; function declaration
    '("\\<\\(function\\)\\s-+&?\\(\\(?:\\sw\\|\\s_\\)+\\)\\s-*("
      (1 font-lock-keyword-face)
      (2 font-lock-function-name-face nil t))

    ;; class hierarchy
    '("\\(self\\|parent\\)\\W" (1 font-lock-constant-face nil nil))

    ;; method and variable features
    '("\\<\\(private\\|protected\\|public\\)\\s-+\\$?\\(?:\\sw\\|\\s_\\)+"
      (1 font-lock-keyword-face))

    ;; method features
    '("^[ \t]*\\(abstract\\|static\\|final\\)\\s-+\\$?\\(?:\\sw\\|\\s_\\)+"
      (1 font-lock-keyword-face))

    ;; variable features
    '("^[ \t]*\\(static\\|const\\)\\s-+\\$?\\(?:\\sw\\|\\s_\\)+"
      (1 font-lock-keyword-face))
    ))
  "Medium level highlighting for PHP mode.")

(defconst php-font-lock-keywords-3
  (append
   php-font-lock-keywords-2
   (list

    ;; <word> or </word> for HTML
    '("</?\\sw+[^>]*>" . font-lock-constant-face)

    ;; HTML entities
    '("&\\w+;" . font-lock-variable-name-face)

    ;; warn about '$' immediately after ->
    '("\\$\\(?:\\sw\\|\\s_\\)+->\\s-*\\(\\$\\)\\(\\(?:\\sw\\|\\s_\\)+\\)"
      (1 font-lock-warning-face) (2 default))

    ;; warn about $word.word -- it could be a valid concatenation,
    ;; but without any spaces we'll assume $word->word was meant.
    '("\\$\\(?:\\sw\\|\\s_\\)+\\(\\.\\)\\sw"
      1 font-lock-warning-face)

    ;; Warn about ==> instead of =>
    '("==+>" . font-lock-warning-face)

    ;; exclude casts from bare-word treatment (may contain spaces)
    `(,(concat "(\\s-*\\(" php-types-regexp "\\)\\s-*)")
      1 font-lock-type-face)

    ;; PHP5: function declarations may contain classes as parameters type
    `(,(concat "[(,]\\s-*\\(\\(?:\\sw\\|\\s_\\)+\\)\\s-+\\$\\(?:\\sw\\|\\s_\\)+\\>")
      1 font-lock-type-face)

    ;; Fontify variables and function calls
    '("\\$\\(this\\|that\\)\\W" (1 font-lock-constant-face nil nil))
    `(,(concat "\\$\\(" php-superglobals-regexp "\\)\\W")
      (1 font-lock-constant-face nil nil)) ; $_GET & co
    '("\\$\\(\\(?:\\sw\\|\\s_\\)+\\)" (1 font-lock-variable-name-face)) ; $variable
    '("->\\(\\(?:\\sw\\|\\s_\\)+\\)" (1 font-lock-variable-name-face t t)) ; ->variable
    '("->\\(\\(?:\\sw\\|\\s_\\)+\\)\\s-*(" . (1 default t t)) ; ->function_call
    '("\\(?:\\sw\\|\\s_\\)+::\\(?:\\sw\\|\\s_\\)+\\s-*(" . default) ; class::method call
    '("\\<\\(?:\\sw\\|\\s_\\)+\\s-*[[(]" . default)	; word( or word[
    '("\\<[0-9]+" . default)		; number (also matches word)

    ;; Warn on any words not already fontified
    '("\\<\\(?:\\sw\\|\\s_\\)+\\>" . font-lock-warning-face)
    ))
  "Gauchy level highlighting for PHP mode.")
  
(defconst php-font-lock-syntactic-keywords
  (if xemacsp nil
    ;; Mark shell-style comments.  font-lock handles this in a
    ;; separate pass from normal syntactic scanning (somehow), so we
    ;; get a chance to mark these in addition to C and C++ style
    ;; comments.  This only works in GNU Emacs, not XEmacs 21 which
    ;; seems to ignore this same code if we try to use it.
    (list
     ;; Mark _all_ # chars as being comment-start.  That will be
     ;; ignored when inside a quoted string.
     '("\\(\#\\)"
       (1 (11 . nil)))
     ;; Mark all newlines ending a line with # as being comment-end.
     ;; This causes a problem, premature end-of-comment, when '#'
     ;; appears inside a multiline C-style comment.  Oh well.
     '("#.*\\([\n]\\)"
       (1 (12 . nil)))
     )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Abbrev ook bindings

(defvar php-mode-abbrev-table nil
  "Abbrev table to use in `php-mode' buffers.")

(defun php-mode-abbrev-table-init ()
  "Initialize `php-mode-abbrev-table'."
  (when php-mode-abbrev-table (clear-abbrev-table php-mode-abbrev-table))
  (define-abbrev-table 'php-mode-abbrev-table
    (append
     '(
       ("break" "" php-template-break-hook 0)
       ("continue" "" php-template-continue-hook 0)
       ("do" "" php-template-do-while-hook 0)
       ("else" "" php-template-else-hook 0)
       ("elseif" "" php-template-elseif-hook 0)
       ("for" "" php-template-for-hook 0)
       ("foreach" "" php-template-foreach-hook 0)
       ("if" "" php-template-if-hook 0)
       ("include" "" php-template-include-hook 0)
       ("include_once" "" php-template-include-once-hook 0)
       ("require" "" php-template-require-hook 0)
       ("require_once" "" php-template-require-once-hook 0)
       ("return" "" php-template-return-hook 0)
       ("switch" "" php-template-switch-hook 0)
       ("while" "" php-template-while-hook 0)
       ("basename" "" php-template-basename-hook 0)
       ("chgrp" "" php-template-chgrp-hook 0)
       ("chmod" "" php-template-chmod-hook 0)
       ("chown" "" php-template-chown-hook 0)
       ("clearstatcache" "" php-template-clearstatcache-hook 0)
       ("copy" "" php-template-copy-hook 0)
       ("delete" "" php-template-delete-hook 0)
       ("dirname" "" php-template-dirname-hook 0)
       ("disk_free_space" "" php-template-disk-free-space-hook 0)
       ("disk_total_space" "" php-template-disk-total-space-hook 0)
       ("diskfreespace" "" php-template-disk-free-space-hook 0)
       ("fclose" "" php-template-fclose-hook 0)
       ("feof" "" php-template-feof-hook 0)
       ("fflush" "" php-template-fflush-hook 0)
       ("fgetc" "" php-template-fgetc-hook 0)
       ("fgetcsv" "" php-template-fgetcsv-hook 0)
       ("fgets" "" php-template-fgets-hook 0)
       ("fgetss" "" php-template-fgetss-hook 0)
       ("file_exists" "" php-template-file-exists-hook 0)
       ("file_get_contents" "" php-template-file-get-contents-hook 0)
       ("file_put_contents" "" php-template-file-put-contents-hook 0)
       ("file" "" php-template-file-hook 0)
       ("fileatime" "" php-template-fileatime-hook 0)
       ("filectime" "" php-template-filectime-hook 0)
       ("filegroup" "" php-template-filegroup-hook 0)
       ("fileinode" "" php-template-fileinode-hook 0)
       ("filemtime" "" php-template-filemtime-hook 0)
       ("fileowner" "" php-template-fileowner-hook 0)
       ("fileperms" "" php-template-fileperms-hook 0)
       ("filesize" "" php-template-filesize-hook 0)
       ("filetype" "" php-template-filetype-hook 0)
       ("flock" "" php-template-flock-hook 0)
       ("fnmatch" "" php-template-fnmatch-hook 0)
       ("fopen" "" php-template-fopen-hook 0)
       ("fpassthru" "" php-template-fpassthru-hook 0)
       ("fputcsv" "" php-template-fputcsv-hook 0)
       ("fput" "" php-template-fwrite-hook 0)
       ("fread" "" php-template-fread-hook 0)
       ;("fscanf" "" php-template-fscanf-hook 0)
       ("fseek" "" php-template-fseek-hook 0)
       ("fstat" "" php-template-fstat-hook 0)
       ("ftell" "" php-template-ftell-hook 0)
       ("ftruncate" "" php-template-ftruncate-hook 0)
       ("fwrite" "" php-template-fwrite-hook 0)
       ("glob" "" php-template-glob-hook 0)
       ("is_dir" "" php-template-is-dir-hook 0)
       ("is_executable" "" php-template-is-executable-hook 0)
       ("is_file" "" php-template-is-file-hook 0)
       ("is_link" "" php-template-is-link-hook 0)
       ("is_readable" "" php-template-is-readable-hook 0)
       ("is_uploaded_file" "" php-template-is-uploaded-file-hook 0)
       ("is_writable" "" php-template-is-writable-hook 0)
       ("is_writeable" "" php-template-is-writable-hook 0)
       ("link" "" php-template-link-hook 0)
       ("linkinfo" "" php-template-linkinfo-hook 0)
       ("lstat" "" php-template-lstat-hook 0)
       ("mkdir" "" php-template-mkdir-hook 0)
       ("move_uploaded_file" "" php-template-move-uploaded-file-hook 0)
       ("parse_ini_file" "" php-template-parse-ini-file-hook 0)
       ("pathinfo" "" php-template-pathinfo-hook 0)
       ("pclose" "" php-template-pclose-hook 0)
       ("popen" "" php-template-popen-hook 0)
       ("readfile" "" php-template-readfile-hook 0)
       ("readlink" "" php-template-readlink-hook 0)
       ("realpath" "" php-template-realpath-hook 0)
       ("rename" "" php-template-rename-hook 0)
       ("rewind" "" php-template-rewind-hook 0)
       ("rmdir" "" php-template-rmdir-hook 0)
       ;("set_file_buffer" "" php-template-set-file-buffer-hook 0)
       ("stat" "" php-template-stat-hook 0)
       ("symlink" "" php-template-symlink-hook 0)
       ("tempnam" "" php-template-tempnam-hook 0)
       ("tmpfile" "" php-template-tmpfile-hook 0)
       ("touch" "" php-template-touch-hook 0)
       ("umask" "" php-template-umask-hook 0)
       ("unlink" "" php-template-unlink-hook 0)
       ("mysql_affected_rows" "" php-template-mysql-affected-rows-hook 0)
       ("mysql_change_user" "" php-template-mysql-change-user-hook 0)
       ("mysql_client_encoding" "" php-template-mysql-client-encoding-hook 0)
       ("mysql_close" "" php-template-mysql-close-hook 0)
       ("mysql_connect" "" php-template-mysql-connect-hook 0)
       ("mysql_create_db" "" php-template-mysql-create-db-hook 0)
       ("mysql_data_seek" "" php-template-mysql-data-seek-hook 0)
       ("mysql_db_name" "" php-template-mysql-db-name-hook 0)
       ("mysql_db_query" "" php-template-mysql-db-query-hook 0)
       ("mysql_drop_db" "" php-template-mysql-drop-db-hook 0)
       ("mysql_errno" "" php-template-mysql-errno-hook 0)
       ("mysql_error" "" php-template-mysql-error-hook 0)
       ("mysql_escape_string" "" php-template-mysql-escape-string-hook 0)
       ("mysql_fetch_array" "" php-template-mysql-fetch-array-hook 0)
       ("mysql_fetch_assoc" "" php-template-mysql-fetch-assoc-hook 0)
       ("mysql_fetch_field" "" php-template-mysql-fetch-field-hook 0)
       ("mysql_fetch_lengths" "" php-template-mysql-fetch-lengths-hook 0)
       ("mysql_fetch_object" "" php-template-mysql-fetch-object-hook 0)
       ("mysql_fetch_row" "" php-template-mysql-fetch-row-hook 0)
       ("mysql_field_flags" "" php-template-mysql-field-flags-hook 0)
       ("mysql_field_len" "" php-template-mysql-field-len-hook 0)
       ("mysql_field_name" "" php-template-mysql-field-name-hook 0)
       ("mysql_field_seek" "" php-template-mysql-field-seek-hook 0)
       ("mysql_field_table" "" php-template-mysql-field-table-hook 0)
       ("mysql_field_type" "" php-template-mysql-field-type-hook 0)
       ("mysql_free_result" "" php-template-mysql-free-result-hook 0)
       ("mysql_get_client_info" "" php-template-mysql-get-client-info-hook 0)
       ("mysql_get_host_info" "" php-template-mysql-get-host-info-hook 0)
       ("mysql_get_proto_info" "" php-template-mysql-get-proto-info-hook 0)
       ("mysql_get_server_info" "" php-template-mysql-get-server-info-hook 0)
       ("mysql_info" "" php-template-mysql-info-hook 0)
       ("mysql_insert_id" "" php-template-mysql-insert-id-hook 0)
       ("mysql_list_dbs" "" php-template-mysql-list-dbs-hook 0)
       ("mysql_list_fields" "" php-template-mysql-list-fields-hook 0)
       ("mysql_list_processes" "" php-template-mysql-list-processes-hook 0)
       ("mysql_list_tables" "" php-template-mysql-list-tables-hook 0)
       ("mysql_num_fields" "" php-template-mysql-num-fields-hook 0)
       ("mysql_num_rows" "" php-template-mysql-num-rows-hook 0)
       ("mysql_pconnect" "" php-template-mysql-pconnect-hook 0)
       ("mysql_ping" "" php-template-mysql-ping-hook 0)
       ("mysql_query" "" php-template-mysql-query-hook 0)
       ("mysql_real_escape_string" "" php-template-mysql-real-escape-string-hook 0)
       ("mysql_result" "" php-template-mysql-result-hook 0)
       ("mysql_select_db" "" php-template-mysql-select-db-hook 0)
       ("mysql_stat" "" php-template-mysql-stat-hook 0)
       ("mysql_tablename" "" php-template-mysql-tablename-hook 0)
       ("mysql_thread_id" "" php-template-mysql-thread-id-hook 0)
       ("mysql_unbuffered_query" "" php-template-mysql-unbuffered-query-hook 0)
       ))))

;; initialize abbrev table for PHP Mode
(php-mode-abbrev-table-init)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Abbrev hooks

(defun php-hooked-abbrev (func)
  "Do function, if syntax says abbrev is a keyword, invoked by hooked abbrev,
but not if inside a comment or quote)."
  (if (php-in-literal)
      (progn
	(insert " ")
	(unexpand-abbrev)
	(delete-char -1))
    (if (not php-electric-mode)
	(progn
	  (insert " ")
	  (unexpand-abbrev)
	  (backward-word 1)
	  (delete-char 1))
      (let ((invoke-char last-command-char)
	    (abbrev-mode -1)
	    (php-template-invoked-by-hook t))
	(let ((caught (catch 'abort
			(funcall func))))
	  (when (stringp caught) (message caught)))
	(when (= invoke-char ?-) (setq abbrev-start-location (point)))
	;; delete CR which is still in event queue
	(if (fboundp 'enqueue-eval-event)
	    (enqueue-eval-event 'delete-char -1)
	  (setq unread-command-events	; push back a delete char
		(list (php-character-to-event ?\177))))))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; PHP-mode

;;;###autoload
(defun php-mode ()
  "Major mode for editing PHP code.

Key bindings:
-------------

\\{php-mode-map}"
  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'php-mode)
  (setq mode-name "PHP")

  ;; set maps and tables
  (use-local-map php-mode-map)
  (set-syntax-table php-mode-syntax-table)
  (setq local-abbrev-table php-mode-abbrev-table)

  ;; set local variables
  (set (make-local-variable 'parse-sexp-ignore-comments) t)
  (set (make-local-variable 'indent-line-function) 'php-indent-line-2)
  (set (make-local-variable 'comment-start) "/* ")
  (set (make-local-variable 'comment-end) " */")
  (set (make-local-variable 'comment-start-skip) "/\\*+ *")
  (set (make-local-variable 'comment-column) 32)
  (set (make-local-variable 'end-comment-column) 70)
  (set (make-local-variable 'comment-indent-function) 'php-comment-indent)
  (set (make-local-variable 'comment-multi-line) t)
  (set (make-local-variable 'indent-tabs-mode) nil)
  (set (make-local-variable 'hippie-expand-verbose) nil)

  (set (make-local-variable 'font-lock-support-mode) 'lazy-lock-mode)
  (set (make-local-variable 'lazy-lock-defer-contextually) nil)
  (set (make-local-variable 'lazy-lock-defer-on-the-fly) t)
  (set (make-local-variable 'lazy-lock-defer-on-scrolling) t)


  ;; add index menu
  (php-index-menu-init)
  ;; add source file menu
  (if php-source-file-menu (php-add-source-files-menu))
  ;; add PHP menu
  (easy-menu-add php-mode-menu-list)
  (easy-menu-define php-mode-menu php-mode-map
		    "Menu keymap for PHP Mode." php-mode-menu-list)

  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults
        '((php-font-lock-keywords-1
           php-font-lock-keywords-2
           ;; Comment-out the next line if the font-coloring is too
           ;; extreme/ugly for you.
           php-font-lock-keywords-3)
          nil                               ; KEYWORDS-ONLY
          T                                 ; CASE-FOLD
          nil                               ; SYNTAX-ALIST
          nil                               ; SYNTAX-BEGIN
          (font-lock-syntactic-keywords . php-font-lock-syntactic-keywords)))

  (php-write-file-hooks-init)
  ;; miscellaneous
  (message "PHP Mode %s.%s" php-version
	   (if noninteractive "" "  See menu for documentation and release notes."))

  ;; run hooks
  (run-hooks 'php-mode-hook))

(defun php-activate-customizations ()
  "Activate all customizations on local variables."
  (interactive)
  (php-mode-map-init)
  (use-local-map php-mode-map)
  (set-syntax-table php-mode-syntax-table)
  (php-update-mode-menu)
  (run-hooks 'menu-bar-update-hook)
  (php-mode-line-update))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Template functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Control Structures

(defun php-template-break ()
  "Insert a break statement."
  (interactive)
  (insert "break;")
  (php-indent-line-2))

(defun php-template-continue ()
  "Insert a continue statement."
  (interactive)
  (insert "continue;")
  (php-indent-line-2))

(defun php-template-do-while ()
  "Insert a do-while statement."
  (interactive)
  (let ((start (point)) here)
    (insert "do {")
    (php-indent-line-2)
    (insert "\n")
    (php-indent-line-2)
    (setq here (point-marker))
    (insert "\n")
    (insert "} while ( ")
    (when (php-template-field "condition" nil t start (point))
      (insert " );")
      (php-indent-line-2)
      (goto-char here))))

(defun php-template-else ()
  "Insert an else statement."
  (interactive)
  (let (here)
    (insert "else {\n")
    (setq here (point-marker))
    (newline-and-indent)
    (insert "}")
    (php-indent-line-2)
    (goto-char here)
    (php-indent-line-2)))

(defun php-template-elseif ()
  "Insert an elseif statement."
  (interactive)
  (let ((start (point)))
    (insert "elseif ")
    (when php-conditions-in-parenthesis (insert "( "))
    (when (php-template-field "condition" nil t start (point))
      (when php-conditions-in-parenthesis (insert " ) {"))
      (newline-and-indent)
      (setq start (point-marker))
      (insert "\n}")
      (php-indent-line-2)
      (goto-char start))))

(defun php-template-for ()
  "Insert a for statement."
  (interactive)
  (let ((start (point)) here)
    (insert "for( ")
    (when (php-template-field "expr1" nil t start (point))
      (insert "; ")
      (php-template-field "expr2" nil t start (point))
      (insert "; ")
      (php-template-field "expr3" nil t start (point))
      (insert " ) {")
      (newline-and-indent)
      (setq here (point-marker))
      (insert "\n}")
      (php-indent-line-2)
      (goto-char here))))

(defun php-template-foreach ()
  "Insert a foreach statement."
  (interactive)
  (let ((start (point)) here)
    (insert "foreach( ")
    (when (php-template-field "array_expression" nil t start (point))
      (insert " as ")
      (php-template-field "value_expression" nil t start (point))
      (insert " ) {")
      (newline-and-indent)
      (setq here (point-marker))
      (insert "\n}")
      (php-indent-line-2)
      (goto-char here))))

(defun php-template-if ()
  "Insert an if statement."
  (interactive)
  (let ((start (point)))
    (insert "if ")
    (when php-conditions-in-parenthesis (insert "( "))
    (when (php-template-field "condition" nil t start (point))
      (when php-conditions-in-parenthesis (insert " ) {"))
      (newline-and-indent)
      (setq start (point-marker))
      (insert "\n}")
      (php-indent-line-2)
      (goto-char start))))

(defun php-template-include-require (label)
  "Insert an include, include_once, require or require_once statement."
  (interactive)
  (let ((start (point)))
    (insert label)
    (if php-include-in-parenthesis 
	(insert "( ")
      (insert " "))
    (when (php-template-field "file" nil t start (point) t "'")
      (if php-include-in-parenthesis 
	  (insert " );")
	(insert ";"))
      (newline-and-indent))))

(defun php-template-include ()
  "Insert an include statement."
  (php-template-include-require "include"))

(defun php-template-include-once ()
  "Insert an include_once statement."
  (php-template-include-require "include_once"))

(defun php-template-require ()
  "Insert a require statement."
  (php-template-include-require "require"))

(defun php-template-require-once ()
  "Insert a require_once statement."
  (php-template-include-require "require_once"))

(defun php-template-return ()
  "Insert a return statement."
  (interactive)
  (let ((start (point)))
    (insert "return ")
    (when (php-template-field "value" nil t start (point))
      (insert ";"))))

(defun php-template-switch ()
  "Insert a switch statement."
  (interactive)
  (let ((start (point)) here)
    (insert "switch( ")
    (when (php-template-field "field" nil t start (point))
      (insert " ) {")
      (newline-and-indent)
      (setq here (point-marker))
      (insert "\n}")
      (php-indent-line-2)
      (goto-char here))))

(defun php-template-while ()
  "Insert a while statement."
  (interactive)
  (let ((start (point)) here)
    (insert "while( ")
    (when (php-template-field "condition" nil t start (point))
      (insert " ) {")
      (newline-and-indent)
      (setq here (point-marker))
      (insert "\n}")
      (php-indent-line-2)
      (goto-char here))))

;; File System

(defun php-template-basename ()
  "Insert a basename statement."
  (interactive)
  (let ((start (point)) remove suffix)
    (insert " = basename( ")
    (setq here (point-marker))
    (goto-char start)
    (when (php-template-field "result_value" nil t start (point) nil "" "$t_basename")
      (goto-char here)
      (when (php-template-field "path" nil t start (point) t "'")
	(setq remove (point))
	(insert ", ")
	(setq suffix (php-template-field "suffix" nil t))
	(if (not suffix)
	    (delete-region remove (point)))
	(insert " );")
	(newline-and-indent)))))

(defun php-template-res-f1-function (label field1)
  "Insert a 'res = label( field1 )' statement (like dirname)."
  (interactive)
  (let ((start (point)) here remove result-value)
    (insert " = ")
    (setq remove (point))
    (insert (concat label "( "))
    (setq here (point-marker))
    (goto-char start)
    (setq result-value (php-template-field "result_value" nil t))
    (if (not result-value)
	(delete-region start remove))
    (goto-char here)
    (when (php-template-field field1 nil t start (point))
      (insert " );")
      (newline-and-indent))))
      
(defun php-template-res-f1-f2-function (label field1 field2)
  "Insert a 'res = label( field1, field2 )' statement (like chgrp, chmod)."
  (interactive)
  (let ((start (point)) here remove result-value)
    (insert " = ")
    (setq remove (point))
    (insert (concat label "( "))
    (setq here (point-marker))
    (goto-char start)
    (setq result-value (php-template-field "result_value" nil t))
    (if (not result-value)
	(delete-region start remove))
    (goto-char here)
    (when (php-template-field field1 nil t start (point))
      (insert ", ")
      (php-template-field field2 nil t)
      (insert " );")
      (newline-and-indent))))
      
(defun php-template-chgrp ()
  "Insert a chgrp statement."
  (interactive)
  (php-template-res-f1-f2-function "chgrp" "filename" "group"))
  
(defun php-template-chmod ()
  "Insert a chgrp statement."
  (interactive)
  (php-template-res-f1-f2-function "chmod" "filename" "mode"))
  
(defun php-template-chown ()
  "Insert a chgrp statement."
  (interactive)
  (php-template-res-f1-f2-function "chown" "filename" "user"))

(defun php-template-void-function (label)
  "Insert a 'void label( void )' function type."
  (interactive)
  (insert label)
  (insert "();")
  (newline-and-indent))

(defun php-template-clearstatcache ()
  "Insert a clearstatcache statement."
  (interactive)
  (php-template-void-function "clearstatcache"))

(defun php-template-copy ()
  "Insert a copy statement."
  (interactive)
  (php-template-res-f1-f2-function "copy" "source" "destination"))

(defun php-template-delete ()
  "Insert a delete statement."
  (interactive)
  (let ((start (point)) here remove)
    (insert "delete( ")
    (when (php-template-field "filename" nil t start (point))
      (insert " );")
      (newline-and-indent))))

(defun php-template-dirname ()
  "Insert a dirname statement."
  (interactive)
  (php-template-res-f1-function "dirname" "path"))

(defun php-template-disk-free-space ()
  "Insert a disk_free_space statement."
  (interactive)
  (php-template-res-f1-function "disk_free_space" "directory"))

(defun php-template-disk-total-space ()
  "Insert a disk_total_space statement."
  (interactive)
  (php-template-res-f1-function "disk_total_space" "directory"))

(defun php-template-search-fopen (&optional include-popen dont-include-fsockopen popen-only)
  "Search about a fopen/fsockopen statement."
  (let (result found)
    (save-excursion
      (if popen-only
	  (setq found (re-search-backward "\$\\(\\w+\\)\\s-*=\\s-*\\popen" nil t))
	(if include-popen
	    (if dont-include-fsockopen
		(setq found (re-search-backward "\$\\(\\w+\\)\\s-*=\\s-*\\(fopen\\|popen\\)" nil t))
	      (setq found (re-search-backward "\$\\(\\w+\\)\\s-*=\\s-*\\(fsockopen\\|fopen\\|popen\\)" nil t)))
	  (if dont-include-fsockopen
	      (setq found (re-search-backward "\$\\(\\w+\\)\\s-*=\\s-*\\fopen" nil t))
	    (setq found (re-search-backward "\$\\(\\w+\\)\\s-*=\\s-*\\(fsockopen\\|fopen\\)" nil t)))))
      (if found
	  (setq result (concat "$" (match-string 1)))
	(setq result nil)))
    result))

(defun php-template-file-generic-function (label field mandatory-count &optional include-popen dont-include-fsockopen popen-only)
  "Generic function template 'result = label( handle, field1, field2...)"
  (interactive)
  (let ((start (point)) remove here result-value elt continue field-count 
	stop file (stop-handle nil) (comment-end (point)))
    (setq file (php-template-search-fopen include-popen dont-include-fsockopen popen-only))
    (when (not file)
      (insert "/* WARNING: ")
      (if popen-only
	  (insert "'popen'")
	(insert "'fopen'")
	(if (not dont-include-fsockopen)
	    (insert ", 'fsockopen'"))
	(if include-popen
	    (insert " or 'popen'")))
      (insert " statement not found in this file */")
      (newline-and-indent)
      (setq comment-end (point)))
    (insert " = ")
    (setq remove (point))
    (insert (concat label "( "))
    (setq here (point-marker))
    (goto-char comment-end)
    (setq result-value (php-template-field "result_value" nil t))
    (if (not result-value)
	(delete-region comment-end remove))
    (goto-char here)
    (if (not file)
	(progn
	  (setq handle (php-template-field "handle" ", "))
	  (if (not handle)
	      (progn (delete-region start (point))
		     (insert (concat label " "))
		     (setq stop-handle t))))
      (insert (concat file ", ")))
    (when (not stop-handle)
      (setq elt field)
      (setq continue t)
      (setq field-count 0)
      (setq stop nil)
      (while (and elt continue)
	(setq result-value (php-template-field (car elt) nil t))
	(if (and (not result-value)
		 (< field-count mandatory-count))
	    (progn (setq continue nil)
		   (delete-region start (point))
		   (insert (concat label " "))
		   (setq stop t))
	  (if (not result-value)
	      (setq continue nil)
	    (insert ", ")))
	(setq field-count (+ 1 field-count))
	(setq elt (cdr elt)))
      (when (not stop)
	(delete-char -2)
	(insert " );")
	(newline-and-indent)))))

(defun php-template-fclose ()
  "Insert a fclose statement."
  (interactive)
  (php-template-file-generic-function "fclose" '() 0))

(defun php-template-feof ()
  "Insert a feof statement."
  (interactive)
  (php-template-file-generic-function "feof" '() 0 t))

(defun php-template-fflush ()
  "Insert a fflush statement."
  (interactive)
  (php-template-file-generic-function "fflush" '() 0 t))

(defun php-template-fgetc ()
  "Insert a fgetc statement."
  (interactive)
  (php-template-file-generic-function "fgetc" '() 0 t))

(defun php-template-fgetcsv ()
  "Insert a fgetcsv statement."
  (interactive)
  (php-template-file-generic-function "fgetcsv" '("length" "delimiter" "enclosure") 0 t))

(defun php-template-fgets ()
  "Insert a fgets statement."
  (interactive)
  (php-template-file-generic-function "fgets" '("length") 0))

(defun php-template-fgetss ()
  "Insert a fgetss statement."
  (interactive)
  (php-template-file-generic-function "fgetss" '("length" "allowable_tags") 0))

(defun php-template-file-exists ()
  "Insert a file_exists statements."
  (interactive)
  (php-template-res-f1-function "file_exists" "filename"))

(defun php-template-generic-function (label field mandatory-count)
  "Generic function template 'result = label( field1, field2...)"
  (interactive)
  (let ((start (point)) remove here result-value elt continue field-count stop)
    (insert " = ")
    (setq remove (point))
    (insert (concat label "( "))
    (setq here (point-marker))
    (goto-char start)
    (setq result-value (php-template-field "result_value" nil t))
    (if (not result-value)
	(delete-region start remove))
    (goto-char here)
    (setq elt field)
    (setq continue t)
    (setq field-count 0)
    (setq stop nil)
    (while (and elt continue)
      (setq result-value (php-template-field (car elt) nil t))
      (if (and (not result-value)
	       (< field-count mandatory-count))
	  (progn (setq continue nil)
		 (delete-region start (point))
		 (insert (concat label " "))
		 (setq stop t))
	(if (not result-value)
	    (setq continue nil)
	  (insert ", ")))
      (setq field-count (+ 1 field-count))
      (setq elt (cdr elt)))
    (when (not stop)
      (if (= 1 field-count)
	  (delete-char -1)
	(delete-char -2)
	(insert " "))
      (insert ");")
      (newline-and-indent))))

(defun php-template-file-get-contents ()
  "Insert a file_get_contents statement."
  (interactive)
  (php-template-generic-function "file_get_contents" '("filename" "use_include_path" "context" "offset" "maxlen") 1))

(defun php-template-file-put-contents ()
  "Insert a file_put_contents statement."
  (interactive)
  (php-template-generic-function "file_put_contents" '("filename" "data" "flags" "context") 2))

(defun php-template-file ()
  "Insert a file statement."
  (interactive)
  (php-template-generic-function "file" '("filename" "use_include_path" "context") 1))

(defun php-template-fileatime ()
  "Insert a fileatime statement."
  (interactive)
  (php-template-generic-function "fileatime" '("filename") 1))

(defun php-template-filectime ()
  "Insert a fileatime statement."
  (interactive)
  (php-template-generic-function "fileatime" '("filename") 1))

(defun php-template-filegroup ()
  "Insert a filegroup statement."
  (interactive)
  (php-template-generic-function "filegroup" '("filename") 1))

(defun php-template-fileinode ()
  "Insert a fileinode statement."
  (interactive)
  (php-template-generic-function "fileinode" '("filename") 1))

(defun php-template-filemtime ()
  "Insert a filemtime statement."
  (interactive)
  (php-template-generic-function "filemtime" '("filename") 1))

(defun php-template-fileowner ()
  "Insert a fileowner statement."
  (interactive)
  (php-template-generic-function "fileowner" '("filename") 1))

(defun php-template-fileperms ()
  "Insert a fileperms statement."
  (interactive)
  (php-template-generic-function "fileperms" '("filename") 1))

(defun php-template-filesize ()
  "Insert a filesize statement."
  (interactive)
  (php-template-generic-function "filesize" '("filename") 1))

(defun php-template-filetype ()
  "Insert a filetype statement."
  (interactive)
  (php-template-generic-function "filetype" '("filename") 1))

(defun php-template-flock ()
  "Insert a flock statement."
  (interactive)
  (php-template-file-generic-function "flock" '("LOCK_SH | LOCK_EX | LOCK_UN | LOCK_NB" "wouldblock") 1 nil t))

(defun php-template-fnmatch ()
  "Insert a fnmatch statement."
  (interactive)
  (php-template-generic-function "fnmatch" '("pattern" "string" "flags") 2))

(defun php-template-fopen ()
  "Inser an fopen statement."
  (interactive)
  (php-template-generic-function "fopen" '("filename" "mode" "use_include_path" "zcontext") 2)
  (when php-add-fclose-with-fopen
    (save-excursion
      (newline-and-indent)
      (php-template-fclose))))

(defun php-template-fpassthru ()
  "Insert a fpassthru statement."
  (interactive)
  (php-template-file-generic-function "fpassthru" '() 0 nil t))

(defun php-template-fputcsv ()
  "Insert a fputcsv statement."
  (interactive)
  (php-template-file-generic-function "fputcsv" '("fields" "delimiter" "enclosure") 0 t))

(defun php-template-fread ()
  "Insert a fread statement."
  (interactive)
  (php-template-file-generic-function "fread" '("length") 1 t))

;(defun php-template-fscanf ()
;  "Insert a fscanf statement.")

(defun php-template-fseek ()
  "Insert a fseek statement."
  (interactive)
  (php-template-file-generic-function "fseek" '("offset" "whence") 1 nil t))

(defun php-template-fstat ()
  "Insert a fstat statement."
  (interactive)
  (php-template-file-generic-function "fstat" '() 0 nil t))

(defun php-template-ftell ()
  "Insert a ftell statement."
  (interactive)
  (php-template-file-generic-function "ftell" '() 0 nil t))

(defun php-template-ftruncate ()
  "Insert a ftruncate statement."
  (interactive)
  (php-template-file-generic-function "ftruncate" '("size") 1 nil t))

(defun php-template-fwrite ()
  "Insert a fwrite statement."
  (interactive)
  (php-template-file-generic-function "fwrite" '("string" "length") 1 nil t))

(defun php-template-glob ()
  "Insert a glob statement."
  (interactive)
  (php-template-generic-function "glob" '("pattern" "flags") 1))

(defun php-template-is-dir ()
  "Insert an is_dir statement."
  (interactive)
  (php-template-generic-function "is_dir" '("filename") 1))

(defun php-template-is-executable ()
  "Insert an is_executable statement."
  (interactive)
  (php-template-generic-function "is_executable" '("filename") 1))

(defun php-template-is-file ()
  "Insert an is_file statement."
  (interactive)
  (php-template-generic-function "is_file" '("filename") 1))

(defun php-template-is-link ()
  "Insert an is_link statement."
  (interactive)
  (php-template-generic-function "is_link" '("filename") 1))

(defun php-template-is-readable ()
  "Insert an is_readable statement."
  (interactive)
  (php-template-generic-function "is_readable" '("filename") 1))

(defun php-template-is-uploaded-file ()
  "Insert an is_uploaded_file statement."
  (interactive)
  (php-template-generic-function "is_uploaded_file" '("filename") 1))

(defun php-template-is-writable ()
  "Insert an is_writable statement."
  (interactive)
  (php-template-generic-function "is_writable" '("filename") 1))

(defun php-template-link ()
  "Insert a link statement."
  (interactive)
  (php-template-generic-function "link" '("target" "link") 2))

(defun php-template-linkinfo ()
  "Insert a linkinfo statement."
  (interactive)
  (php-template-generic-function "linkinfo" '("path") 1))

(defun php-template-lstat ()
  "Insert a lstat statement."
  (interactive)
  (php-template-generic-function "lstat" '("filename") 1))

(defun php-template-mkdir ()
  "Insert a mkdir statement."
  (interactive)
  (php-template-generic-function "mkdir" '("pathname" "mode" "recursive" "context") 1))

(defun php-template-move-uploaded-file ()
  "Insert a move_uploaded_file statement."
  (interactive)
  (php-template-generic-function "move_uploaded_file" '("filename" "destination") 2))

(defun php-template-parse-ini-file ()
  "Insert a parse_ini_file statement."
  (interactive)
  (php-template-generic-function "parse_ini_file" '("filename" "process_sections") 1))

(defun php-template-pathinfo ()
  "Insert a pathinfo statement."
  (interactive)
  (php-template-generic-function "pathinfo" '("path" "options") 1))

(defun php-template-pclose ()
  "Insert a pclose statement."
  (interactive)
  (php-template-file-generic-function "pclose" '() 0 nil nil t))

(defun php-template-popen ()
  "Inser a popen statement."
  (interactive)
  (php-template-generic-function "popen" '("command" "mode") 2)
  (when php-add-fclose-with-fopen
    (save-excursion
      (newline-and-indent)
      (php-template-pclose))))

(defun php-template-readfile ()
  "Insert a readfile statement."
  (interactive)
  (php-template-generic-function "readfile" '("filename" "use_include_path" "context") 1))

(defun php-template-readlink ()
  "Insert a readlink statement."
  (interactive)
  (php-template-generic-function "readlink" '("path") 0))

(defun php-template-realpath ()
  "Insert a realpath statement."
  (interactive)
  (php-template-generic-function "realpath" '("path") 1))

(defun php-template-rename ()
  "Insert a rename statement."
  (interactive)
  (php-template-generic-function "rename" '("oldname" "newname" "context") 2))

(defun php-template-rewind ()
  "Insert a rewind statement."
  (interactive)
  (php-template-file-generic-function "rewind" '() 0 nil t))

(defun php-template-rmdir ()
  "Insert a rmdir statement."
  (interactive)
  (php-template-generic-function "rmdir" '("dirname" "context") 1))

(defun php-template-stat ()
  "Insert a stat statement."
  (interactive)
  (php-template-generic-function "stat" '("filename") 1))

(defun php-template-symlink ()
  "Insert a symlink statement."
  (interactive)
  (php-template-generic-function "symlink" '("target" "link") 2))

(defun php-template-tempnam ()
  "Insert a tempnam statement."
  (interactive)
  (php-template-generic-function "tempnam" '("dir" "prefix") 2))

(defun php-template-tmpfilet ()
  "Insert a tmpfile statement."
  (interactive)
  (php-template-generic-function "tmpfile" '() 0))

(defun php-template-touch ()
  "Insert a touch statement."
  (interactive)
  (php-template-generic-function "touch" '("filename" "time" "atime") 1))

(defun php-template-umask ()
  "Insert an umask statement."
  (interactive)
  (php-template-generic-function "umask" '("mask") 0))

(defun php-template-unlink ()
  "Insert an unlink statement."
  (interactive)
  (php-template-generic-function "unlink" '("filename" "context") 1))

;; MySQL

(defun php-template-search-mysql-connect ()
  "Search about a mysl_connect statement."
  (let (result found)
    (save-excursion
      (setq found (re-search-backward "\$\\(\\w+\\)\\s-*=\\s-*\\(mysql_connect\\|mysql_pconnect\\)" nil t))
      (if found
	  (setq result (concat "$" (match-string 1)))
	(setq result nil)))
    result))

(defun php-template-search-mysql-query (query)
  "Search about a mysl_query statement."
  (let (result found)
    (save-excursion
      (setq found (re-search-backward (concat "\$\\(\\w+\\)\\s-*=\\s-*\\" query) nil t))
      (if found
	  (setq result (concat "$" (match-string 1)))
	(setq result nil)))
    result))

(defun php-template-mysql-generic-function (label field mandatory-count)
  "Generic function for MySQL functions."
  (interactive)
  (let ((start (point)) remove here result-value elt continue field-count 
	stop file (stop-handle nil) (comment-end (point)))
    (setq file (php-template-search-mysql-connect))
    (insert " = ")
    (setq remove (point))
    (insert (concat label "( "))
    (setq here (point-marker))
    (goto-char comment-end)
    (setq result-value (php-template-field "result_value" nil t))
    (if (not result-value)
	(delete-region comment-end remove))
    (goto-char here)
    (setq elt field)
    (setq continue t)
    (setq field-count 0)
    (setq stop nil)
    (while (and elt continue)
      (setq result-value (php-template-field (car elt) nil t))
      (if (and (not result-value)
	       (< field-count mandatory-count))
	  (progn (setq continue nil)
		 (delete-region start (point))
		 (insert (concat label " "))
		 (setq stop t))
	(if (not result-value)
	    (setq continue nil)
	  (insert ", ")))
      (setq field-count (+ 1 field-count))
      (setq elt (cdr elt)))
    (when (not stop)
      (if (and (not file) continue)
	  (progn
	    (setq here (point-marker))
	    (setq handle (php-template-field "link_identifier" "  "))
	    (if (not handle)
		(progn (delete-region here (point))
		       (insert " ")
		       (setq stop-handle t))))
	(if continue
	    (insert (concat file "  "))))
      (delete-char -2)
      (if (= (preceding-char) ?\,)
	  (delete-char -1))
      (if (= (preceding-char) ?\()
	  (insert ");")
	(insert " );"))
      (newline-and-indent))))

(defun php-template-mysql-result-generic-function (query label field mandatory-count)
  "Generic function for MySQL functions."
  (interactive)
  (let ((start (point)) remove here result-value elt continue field-count 
	stop file (stop-handle nil) (comment-end (point)))
    (setq file (php-template-search-mysql-query query))
    (when (not file)
      (insert "/* WARNING: '")
      (insert query)
      (insert "' statement not found in this file */")
      (newline-and-indent)
      (setq comment-end (point)))
    (insert " = ")
    (setq remove (point))
    (insert (concat label "( "))
    (setq here (point-marker))
    (goto-char comment-end)
    (setq result-value (php-template-field "result_value" nil t))
    (if (not result-value)
	(delete-region comment-end remove))
    (goto-char here)
    (setq handle (php-template-field "result" ", " nil (point) (point) nil "" file))
    (if (not handle)
	(progn (delete-region start (point))
	       (insert (concat label " "))
	       (setq stop-handle t)))
    (when (not stop-handle)
      (setq elt field)
      (setq continue t)
      (setq field-count 0)
      (setq stop nil)
      (while (and elt continue)
	(setq result-value (php-template-field (car elt) nil t))
	(if (and (not result-value)
		 (< field-count mandatory-count))
	    (progn (setq continue nil)
		   (delete-region start (point))
		   (insert (concat label " "))
		   (setq stop t))
	  (if (not result-value)
	      (setq continue nil)
	    (insert ", ")))
	(setq field-count (+ 1 field-count))
	(setq elt (cdr elt)))
      (when (not stop)
	(delete-char -2)
	(insert " );")
	(newline-and-indent)))))

(defun php-template-mysql-affected-rows ()
  "Insert a mysql_affected_rows statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_affected_rows" '() 0))

(defun php-template-mysql-change-user ()
  "Insert a mysql_change_user statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_change_user" '("user" "password" "database") 2))

(defun php-template-mysql-client-encoding ()
  "Insert a mysql_client_encoding statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_client_encoding" '() 0))

(defun php-template-mysql-close ()
  "Insert a mysql_close statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_close" '() 0))

(defun php-template-mysql-connect ()
  "Insert a mysql_connect statement."
  (interactive)
  (php-template-generic-function "mysql_connect" '("server" "username" "password" "new_link" "client_flags") 0)
  (when php-add-mysql-close-when-connect
    (save-excursion
      (newline-and-indent)
      (php-template-mysql-close))))

(defun php-template-mysql-create-db ()
  "Insert a mysql_create_db statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_create_db" '("database_name") 1))

(defun php-template-mysql-data-seek ()
  "Insert a mysql_data_seek statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_data_seek" '("row_number") 1))

(defun php-template-mysql-db-name ()
  "Insert a mysql_db_name statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_list_dbs" "mysql_db_name" '("row" "field") 1))

(defun php-template-mysql-db-query ()
  "Insert a mysql_db_query statement."
  (interactive)
  (php-template-mysql-select-db)
  (php-template-mysql-query))

(defun php-template-mysql-drop-db ()
  "Insert a mysql_drop_db statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_drop_db" '("database_name") 1))

(defun php-template-mysql-errno ()
  "Insert a mysql_errno statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_errno" '() 0))

(defun php-template-mysql-error ()
  "Insert a mysql_error statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_error" '() 0))

(defun php-template-mysql-escape-string ()
  "Insert a mysql_escape_string statement."
  (interactive)
  (php-template-generic-function "mysql_escape_string" '("unescaped_string") 1))

(defun php-template-mysql-fetch-array ()
  "Insert a mysql_fetch_array statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_fetch_array" '("result_type") 0))

(defun php-template-mysql-fetch-assoc ()
  "Insert a mysql_fetch_assoc statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_fetch_assoc" '() 0))

(defun php-template-mysql-fetch-field ()
  "Insert a mysql_fetch_field statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_fetch_field" '("field_offset") 0))

(defun php-template-mysql-fetch-lengths ()
  "Insert a mysql_fetch_lengths statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_fetch_lengths" '() 0))

(defun php-template-mysql-fetch-object ()
  "Insert a mysql_fetch_object statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_fetch_object" '() 0))

(defun php-template-mysql-fetch-row ()
  "Insert a mysql_fetch_row statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_fetch_row" '() 0))

(defun php-template-mysql-field-flags ()
  "Insert a mysql_field_flags statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_field_flags" '("field_offset") 1))

(defun php-template-mysql-field-len ()
  "Insert a mysql_field_len statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_field_len" '("field_offset") 1))

(defun php-template-mysql-field-name ()
  "Insert a mysql_field_name statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_field_name" '("field_offset") 1))

(defun php-template-mysql-field-seeks ()
  "Insert a mysql_field_seek statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_field_seek" '("field_offset") 1))

(defun php-template-mysql-field-table ()
  "Insert a mysql_field_table statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_field_table" '("field_offset") 1))

(defun php-template-mysql-field-type ()
  "Insert a mysql_field_type statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_field_type" '("field_offset") 1))

(defun php-template-mysql-free-result ()
  "Insert a mysql_free_result statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_free_result" '() 0))

(defun php-template-mysql-get-client-info ()
  "Insert a mysql_get_client_info statement."
  (interactive)
  (php-template-generic-function "mysql_get_client_info" '() 0))

(defun php-template-mysql-get-host-info ()
  "Insert a mysql_get_host_info statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_get_host_info" '() 0))

(defun php-template-mysql-get-proto-info ()
  "Insert a mysql_get_ptoto_info statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_get_proto_info" '() 0))

(defun php-template-mysql-get-server-info ()
  "Insert a mysql_get_server_info statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_get_server_info" '() 0))

(defun php-template-mysql-info ()
  "Insert a mysql_info statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_info" '() 0))

(defun php-template-mysql-list-dbs ()
  "Insert a mysql_list_dbs statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_list_dbs" '() 0))

(defun php-template-mysql-list-fields ()
  "Insert a mysql_list_fields statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_list_fields" '("database_name" "table_name") 2))

(defun php-template-mysql-list-processes ()
  "Insert a mysql_list_processes statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_list_processes" '() 0))

(defun php-template-mysql-list-tables ()
  "Insert a mysql_list_tables statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_list_tables" '("database") 1))

(defun php-template-mysql-num-fields ()
  "Insert a mysql_num_fields statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_num_fields" '() 0))

(defun php-template-mysql-num-rows ()
  "Insert a mysql_num_rows statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_num_rows" '() 0))

(defun php-template-mysql-pconnect ()
  "Insert a mysql_pconnect statement."
  (interactive)
  (php-template-generic-function "mysql_pconnect" '("server" "username" "password" "client_flags") 0)
  (when php-add-mysql-close-when-connect
    (save-excursion
      (newline-and-indent)
      (php-template-mysql-close))))

(defun php-template-mysql-ping ()
  "Insert a mysql_ping statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_ping" '() 0))

(defun php-template-mysql-query ()
  "Insert a mysql_query statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_query" '("query") 1))

(defun php-template-mysql-real-escape-string ()
  "Insert a mysql_real_escape_string statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_real_escape_string" '("unescaped_string") 1))

(defun php-template-mysql-result ()
  "Insert a mysql_result statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_query" "mysql_result" '("row" "field") 1))

(defun php-template-mysql-select-db ()
  "Insert a mysql_select_db statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_select_db" '("database_name") 1))

(defun php-template-mysql-stat ()
  "Insert a mysql_stat statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_stat" '() 0))

(defun php-template-mysql-tablename ()
  "Insert a mysql_tablename statement."
  (interactive)
  (php-template-mysql-result-generic-function "mysql_list_tables" "mysql_tablename" '("i") 1))

(defun php-template-mysql-thread-id ()
  "Insert a mysql_thread_id statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_thread_id" '() 0))

(defun php-template-mysql-unbuffered-query ()
  "Insert a mysql_unbuffered_query statement."
  (interactive)
  (php-template-mysql-generic-function "mysql_unbuffered_query" '("query") 1))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Template hooks
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Control Structures
(defun php-template-break-hook ()
  (php-hooked-abbrev 'php-template-beak))
(defun php-template-continue-hook ()
  (php-hooked-abbrev 'php-template-continue))
(defun php-template-do-while-hook ()
  (php-hooked-abbrev 'php-template-do-while))
(defun php-template-else-hook ()
  (php-hooked-abbrev 'php-template-else))
(defun php-template-elseif-hook ()
  (php-hooked-abbrev 'php-template-elseif))
(defun php-template-for-hook ()
  (php-hooked-abbrev 'php-template-for))
(defun php-template-foreach-hook ()
  (php-hooked-abbrev 'php-template-foreach))
(defun php-template-if-hook ()
  (php-hooked-abbrev 'php-template-if))
(defun php-template-include-hook ()
  (php-hooked-abbrev 'php-template-include))
(defun php-template-include-once-hook ()
  (php-hooked-abbrev 'php-template-include-once))
(defun php-template-require-hook ()
  (php-hooked-abbrev 'php-template-require))
(defun php-template-require-once-hook ()
  (php-hooked-abbrev 'php-template-require-once))
(defun php-template-return-hook ()
  (php-hooked-abbrev 'php-template-return))
(defun php-template-switch-hook ()
  (php-hooked-abbrev 'php-template-switch))
(defun php-template-while-hook ()
  (php-hooked-abbrev 'php-template-while))

;; File System
(defun php-template-basename-hook ()
  (php-hooked-abbrev 'php-template-basename))
(defun php-template-chgrp-hook ()
  (php-hooked-abbrev 'php-template-chgrp))
(defun php-template-chmod-hook ()
  (php-hooked-abbrev 'php-template-chmod))
(defun php-template-chown-hook ()
  (php-hooked-abbrev 'php-template-chown))
(defun php-template-clearstatcache-hook ()
  (php-hooked-abbrev 'php-template-clearstatcache))
(defun php-template-copy-hook ()
  (php-hooked-abbrev 'php-template-copy))
(defun php-template-delete-hook ()
  (php-hooked-abbrev 'php-template-delete))
(defun php-template-dirname-hook ()
  (php-hooked-abbrev 'php-template-dirname))
(defun php-template-disk-free-space-hook ()
  (php-hooked-abbrev 'php-template-disk-free-space))
(defun php-template-disk-total-space-hook ()
  (php-hooked-abbrev 'php-template-disk-total-space))
(defun php-template-fclose-hook ()
  (php-hooked-abbrev 'php-template-fclose))
(defun php-template-feof-hook ()
  (php-hooked-abbrev 'php-template-feof))
(defun php-template-fflush-hook ()
  (php-hooked-abbrev 'php-template-fflush))
(defun php-template-fgetc-hook ()
  (php-hooked-abbrev 'php-template-fgetc))
(defun php-template-fgetcsv-hook ()
  (php-hooked-abbrev 'php-template-fgetcsv))
(defun php-template-fgets-hook ()
  (php-hooked-abbrev 'php-template-fgets))
(defun php-template-fgetss-hook ()
  (php-hooked-abbrev 'php-template-fgetss))
(defun php-template-file-exists-hook ()
  (php-hooked-abbrev 'php-template-file-exists))
(defun php-template-file-get-contents-hook ()
  (php-hooked-abbrev 'php-template-file-get-contents))
(defun php-template-file-put-contents-hook ()
  (php-hooked-abbrev 'php-template-file-put-contents))
(defun php-template-file-hook ()
  (php-hooked-abbrev 'php-template-file))
(defun php-template-fileatime-hook ()
  (php-hooked-abbrev 'php-template-fileatime))
(defun php-template-filectime-hook ()
  (php-hooked-abbrev 'php-template-filectime))
(defun php-template-filegroup-hook ()
  (php-hooked-abbrev 'php-template-filegroup))
(defun php-template-fileinode-hook ()
  (php-hooked-abbrev 'php-template-fileinode))
(defun php-template-filemtime-hook ()
  (php-hooked-abbrev 'php-template-filemtime))
(defun php-template-fileowner-hook ()
  (php-hooked-abbrev 'php-template-fileowner))
(defun php-template-fileperms-hook ()
  (php-hooked-abbrev 'php-template-fileperms))
(defun php-template-filesize-hook ()
  (php-hooked-abbrev 'php-template-filesize))
(defun php-template-filetype-hook ()
  (php-hooked-abbrev 'php-template-filetype))
(defun php-template-flock-hook ()
  (php-hooked-abbrev 'php-template-flock))
(defun php-template-fnmatch-hook ()
  (php-hooked-abbrev 'php-template-fnmatch))
(defun php-template-fopen-hook ()
  (php-hooked-abbrev 'php-template-fopen))
(defun php-template-fpassthru-hook ()
  (php-hooked-abbrev 'php-template-fpassthru))
(defun php-template-fputcsv-hook ()
  (php-hooked-abbrev 'php-template-fputcsv))
(defun php-template-fread-hook ()
  (php-hooked-abbrev 'php-template-fread))
(defun php-template-fscanf-hook ()
  (php-hooked-abbrev 'php-template-fscanf))
(defun php-template-fseek-hook ()
  (php-hooked-abbrev 'php-template-fseek))
(defun php-template-fstat-hook ()
  (php-hooked-abbrev 'php-template-fstat))
(defun php-template-ftell-hook ()
  (php-hooked-abbrev 'php-template-ftell))
(defun php-template-ftruncate-hook ()
  (php-hooked-abbrev 'php-template-ftruncate))
(defun php-template-fwrite-hook ()
  (php-hooked-abbrev 'php-template-fwrite))
(defun php-template-glob-hook ()
  (php-hooked-abbrev 'php-template-glob))
(defun php-template-is-dir-hook ()
  (php-hooked-abbrev 'php-template-is-dir))
(defun php-template-is-executable-hook ()
  (php-hooked-abbrev 'php-template-is-executable))
(defun php-template-is-file-hook ()
  (php-hooked-abbrev 'php-template-is-file))
(defun php-template-is-link-hook ()
  (php-hooked-abbrev 'php-template-is-link))
(defun php-template-is-readable-hook ()
  (php-hooked-abbrev 'php-template-is-readable))
(defun php-template-is-uploaded-file-hook ()
  (php-hooked-abbrev 'php-template-is-uploaded-file))
(defun php-template-is-writable-hook ()
  (php-hooked-abbrev 'php-template-is-writable))
(defun php-template-link-hook ()
  (php-hooked-abbrev 'php-template-link))
(defun php-template-linkinfo-hook ()
  (php-hooked-abbrev 'php-template-linkinfo))
(defun php-template-lstat-hook ()
  (php-hooked-abbrev 'php-template-lstat))
(defun php-template-mkdir-hook ()
  (php-hooked-abbrev 'php-template-mkdir))
(defun php-template-move-uploaded-file-hook ()
  (php-hooked-abbrev 'php-template-move-uploaded-file))
(defun php-template-parse-ini-file-hook ()
  (php-hooked-abbrev 'php-template-parse-ini-file))
(defun php-template-pathinfo-hook ()
  (php-hooked-abbrev 'php-template-pathinfo))
(defun php-template-pclose-hook ()
  (php-hooked-abbrev 'php-template-pclose))
(defun php-template-popen-hook ()
  (php-hooked-abbrev 'php-template-popen))
(defun php-template-readfile-hook ()
  (php-hooked-abbrev 'php-template-readfile))
(defun php-template-readlink-hook ()
  (php-hooked-abbrev 'php-template-readlink))
(defun php-template-realpath-hook ()
  (php-hooked-abbrev 'php-template-realpath))
(defun php-template-rename-hook ()
  (php-hooked-abbrev 'php-template-rename))
(defun php-template-rewind-hook ()
  (php-hooked-abbrev 'php-template-rewind))
(defun php-template-rmdir-hook ()
  (php-hooked-abbrev 'php-template-rmdir))
(defun php-template-stat-hook ()
  (php-hooked-abbrev 'php-template-stat))
(defun php-template-symlink-hook ()
  (php-hooked-abbrev 'php-template-symlink))
(defun php-template-tempnam-hook ()
  (php-hooked-abbrev 'php-template-tempnam))
(defun php-template-tmpfile-hook ()
  (php-hooked-abbrev 'php-template-tmpfile))
(defun php-template-touch-hook ()
  (php-hooked-abbrev 'php-template-touch))
(defun php-template-umask-hook ()
  (php-hooked-abbrev 'php-template-umask))

;; MySQL
(defun php-template-mysql-affected-rows-hook ()
  (php-hooked-abbrev 'php-template-mysql-affected-rows))
(defun php-template-mysql-change-user-hook ()
  (php-hooked-abbrev 'php-template-mysql-change-user))
(defun php-template-mysql-client-encoding-hook ()
  (php-hooked-abbrev 'php-template-mysql-client-encoding))
(defun php-template-mysql-close-hook ()
  (php-hooked-abbrev 'php-template-mysql-close))
(defun php-template-mysql-connect-hook ()
  (php-hooked-abbrev 'php-template-mysql-connect))
(defun php-template-mysql-create-db-hook ()
  (php-hooked-abbrev 'php-template-mysql-create-db))
(defun php-template-mysql-data-seek-hook ()
  (php-hooked-abbrev 'php-template-mysql-data-seek))
(defun php-template-mysql-db-name-hook ()
  (php-hooked-abbrev 'php-template-mysql-db-name))
(defun php-template-mysql-db-query-hook ()
  (php-hooked-abbrev 'php-template-mysql-db-query))
(defun php-template-mysql-drop-db-hook ()
  (php-hooked-abbrev 'php-template-mysql-drop-db))
(defun php-template-mysql-errno-hook ()
  (php-hooked-abbrev 'php-template-mysql-errno))
(defun php-template-mysql-error-hook ()
  (php-hooked-abbrev 'php-template-mysql-error))
(defun php-template-mysql-escape-string-hook ()
  (php-hooked-abbrev 'php-template-mysql-escape-string))
(defun php-template-mysql-fetch-array-hook ()
  (php-hooked-abbrev 'php-template-mysql-fetch-array))
(defun php-template-mysql-fetch-assoc-hook ()
  (php-hooked-abbrev 'php-template-mysql-fetch-assoc))
(defun php-template-mysql-fetch-field-hook ()
  (php-hooked-abbrev 'php-template-mysql-fetch-field))
(defun php-template-mysql-fetch-lengths-hook ()
  (php-hooked-abbrev 'php-template-mysql-fetch-lengths))
(defun php-template-mysql-fetch-object-hook ()
  (php-hooked-abbrev 'php-template-mysql-fetch-object))
(defun php-template-mysql-fetch-row-hook ()
  (php-hooked-abbrev 'php-template-mysql-fetch-row))
(defun php-template-mysql-field-flags-hook ()
  (php-hooked-abbrev 'php-template-mysql-field-flags))
(defun php-template-mysql-field-len-hook ()
  (php-hooked-abbrev 'php-template-mysql-field-len))
(defun php-template-mysql-field-name-hook ()
  (php-hooked-abbrev 'php-template-mysql-field-name))
(defun php-template-mysql-field-seek-hook ()
  (php-hooked-abbrev 'php-template-mysql-field-seek))
(defun php-template-mysql-field-table-hook ()
  (php-hooked-abbrev 'php-template-mysql-field-table))
(defun php-template-mysql-field-type-hook ()
  (php-hooked-abbrev 'php-template-mysql-field-type))
(defun php-template-mysql-free-result-hook ()
  (php-hooked-abbrev 'php-template-mysql-free-result))
(defun php-template-mysql-get-client-info-hook ()
  (php-hooked-abbrev 'php-template-mysql-get-client-info))
(defun php-template-mysql-get-host-info-hook ()
  (php-hooked-abbrev 'php-template-mysql-get-host-info))
(defun php-template-mysql-get-proto-info-hook ()
  (php-hooked-abbrev 'php-template-mysql-get-proto-info))
(defun php-template-mysql-get-server-info-hook ()
  (php-hooked-abbrev 'php-template-mysql-get-server-info))
(defun php-template-mysql-info-hook ()
  (php-hooked-abbrev 'php-template-mysql-info))
(defun php-template-mysql-insert-id-hook ()
  (php-hooked-abbrev 'php-template-mysql-insert-id))
(defun php-template-mysql-list-dbs-hook ()
  (php-hooked-abbrev 'php-template-mysql-list-dbs))
(defun php-template-mysql-list-fields-hook ()
  (php-hooked-abbrev 'php-template-mysql-list-fields))
(defun php-template-mysql-list-processes-hook ()
  (php-hooked-abbrev 'php-template-mysql-list-processes))
(defun php-template-mysql-list-tables-hook ()
  (php-hooked-abbrev 'php-template-mysql-list-tables))
(defun php-template-mysql-num-fields-hook ()
  (php-hooked-abbrev 'php-template-mysql-num-fields))
(defun php-template-mysql-num-rows-hook ()
  (php-hooked-abbrev 'php-template-mysql-num-rows))
(defun php-template-mysql-pconnect-hook ()
  (php-hooked-abbrev 'php-template-mysql-pconnect))
(defun php-template-mysql-ping-hook ()
  (php-hooked-abbrev 'php-template-mysql-ping))
(defun php-template-mysql-query-hook ()
  (php-hooked-abbrev 'php-template-mysql-query))
(defun php-template-mysql-real-escape-string-hook ()
  (php-hooked-abbrev 'php-template-mysql-real-escape-string))
(defun php-template-mysql-result-hook ()
  (php-hooked-abbrev 'php-template-mysql-result))
(defun php-template-mysql-select-db-hook ()
  (php-hooked-abbrev 'php-template-mysql-select-db))
(defun php-template-mysql-stat-hook ()
  (php-hooked-abbrev 'php-template-mysql-stat))
(defun php-template-mysql-tablename-hook ()
  (php-hooked-abbrev 'php-template-mysql-tablename))
(defun php-template-mysql-thread-id-hook ()
  (php-hooked-abbrev 'php-template-mysql-thread-id))
(defun php-template-mysql-unbuffered-query-hook ()
  (php-hooked-abbrev 'php-template-mysql-unbuffered-query))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Template functions utilities
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun php-template-field (prompt &optional follow-string optional
				   begin end is-string string-char default)
  "Prompt for string and insert it in buffer with optional FOLLOW-STRING.
If OPTIONAL is nil, the prompt is left if an empty string is inserted.  If
an empty string is inserted, return nil and call `php-template-undo' for
the region between BEGIN and END.  IS-STRING indicates whether a string
with double-quotes is to be inserted.  DEFAULT specifies a default string."
  (let ((position (point))
	string)
    (insert "<" prompt ">")
    (if (not (> (length string-char) 0))
	(setq string-char "\""))
    (setq string
	  (condition-case ()
	      (read-from-minibuffer (concat prompt ": ")
				    (or (and is-string (cons (concat string-char string-char) 1)) default)
				    php-minibuffer-local-map)
	    (quit (if (and optional begin end)
		      (progn (beep) "")
		    (keyboard-quit)))))
    (when (or (not (equal string "")) optional)
      (delete-region position (point)))
    (when (and (equal string "") optional begin end)
      (php-template-undo begin end)
      (message "Template aborted"))
    (unless (equal string "")
      (insert string))
    (when (or (not (equal string "")) (not optional))
      (insert (or follow-string "")))
    (if (equal string "") nil string)))

(defun php-template-undo (begin end)
  "Undo aborted template by deleting region and unexpanding the keyword."
  (cond (php-template-invoked-by-hook
	 (goto-char end)
	 (insert " ")
	 (delete-region begin end)
	 (unexpand-abbrev))
	(t (delete-region begin end))))

(defun php-resolve-env-variable (string)
  "Resolve environment variables in STRING."
  (while (string-match "\\(.*\\)${?\\(\\(\\w\\|_\\)+\\)}?\\(.*\\)" string)
    (setq string (concat (match-string 1 string)
			 (getenv (match-string 2 string))
			 (match-string 4 string))))
  string)

(defun php-insert-string-or-file (string)
  "Insert STRING or file contents if STRING is an existing file name."
  (unless (equal string "")
    (let ((file-name
	   (progn (string-match "^\\([^\n]+\\)" string)
		  (php-resolve-env-variable (match-string 1 string)))))
      (if (file-exists-p file-name)
	   (forward-char (cadr (insert-file-contents file-name)))
	(insert string)))))

(defun php-template-modify (&optional noerror)
  "Actualize modification date."
  (interactive)
  (php-prepare-search-2
   (save-excursion
     (goto-char (point-min))
     (if (re-search-forward php-modify-date-prefix-string nil t)
	 (progn (delete-region (point) (progn (end-of-line) (point)))
		(php-template-insert-date))
       (unless noerror
	 (error (concat "ERROR:  Modification date prefix string \""
			php-modify-date-prefix-string "\" not found")))))))

(defun php-template-modify-noerror ()
  "Call `php-template-modify' with NOERROR non-nil."
  (php-template-modify t))

(defun php-template-insert-date ()
  "Insert date in appropriate format."
  (interactive)
  (insert
   (cond
    ;; 'american, 'european, 'scientific kept for backward compatibility
    ((eq php-date-format 'american) (format-time-string "%m/%d/%Y" nil))
    ((eq php-date-format 'european) (format-time-string "%d.%m.%Y" nil))
    ((eq php-date-format 'scientific) (format-time-string "%Y/%m/%d" nil))
    (t (format-time-string php-date-format nil)))))

(defun php-template-header (&optional file-title)
  "Insert a PHP file header."
  (interactive)
  (unless (equal php-file-header "")
    (let (pos)
      (save-excursion
	(php-insert-string-or-file php-file-header)
	(setq pos (point-marker)))
      (php-template-replace-header-keywords
       (point-min-marker) pos file-title))))

(defun php-template-footer ()
  "Insert a PHP file footer."
  (interactive)
  (unless (equal php-file-footer "")
    (let (pos)
      (save-excursion
	(setq pos (point-marker))
	(php-insert-string-or-file php-file-footer)
	(unless (= (preceding-char) ?\n)
	  (insert "\n")))
      (php-template-replace-header-keywords pos (point-max-marker)))))

(defun php-template-replace-header-keywords (beg end &optional file-title is-model)
  "Replace keywords in header and footer."
  (let (pos)
    (php-prepare-search-2
     (save-excursion
       (goto-char beg)
       (while (search-forward "<filename>" end t)
	 (replace-match (buffer-name) t t))
       (goto-char beg)
       (while (search-forward "<copyright>" end t)
	 (replace-match php-copyright-string t t))
       (goto-char beg)
       (while (search-forward "<author>" end t)
	 (replace-match "" t t)
	 (insert (user-full-name))
	 (when user-mail-address (insert "  <" user-mail-address ">")))
       (goto-char beg)
       (while (search-forward "<login>" end t)
	 (replace-match (user-login-name) t t))
       (goto-char beg)
       (while (search-forward "<company>" end t)
	 (replace-match php-company-name t t))
       (goto-char beg)
       ;; Replace <RCS> with $, so that RCS for the source is
       ;; not over-enthusiastic with replacements
       (while (search-forward "<RCS>" end t)
	 (replace-match "$" nil t))
       (goto-char beg)
       (while (search-forward "<date>" end t)
	 (replace-match "" t t)
	 (php-template-insert-date))
       (goto-char beg)
       (while (search-forward "<year>" end t)
	 (replace-match (format-time-string "%Y" nil) t t))
       (goto-char beg)
       (let (string)
	 (while
	     (re-search-forward "<\\(\\(\\w\\|\\s_\\)*\\) string>" end t)
	   (setq string (read-string (concat (match-string 1) ": ")))
	   (replace-match string t t)))
       (goto-char beg)
       (when (and (not is-model) (search-forward "<cursor>" end t))
	 (replace-match "" t t)
	 (setq pos (point)))))))

(provide 'php-mode)
